
def set_template_string_multi(chart_width):

    pos = [16, 9]

    template_string = '''
        <html>
        <head>
        <link rel="stylesheet" type="text/css" href="{{style_sheet}}">

        <script src="https://www.gstatic.com/charts/loader.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/rgbcolor.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/StackBlur.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/canvg.js"></script>
        <script type="text/javascript">

        google.setOnLoadCallback(drawChart);

        function drawChart() {
        var dataTable = new google.visualization.DataTable();
        dataTable.addColumn({ type: 'string', id: 'Core' });
        dataTable.addColumn({ type: 'string', id: 'Message' });
        dataTable.addColumn({ type: 'number', id: 'Start' });
        dataTable.addColumn({ type: 'number', id: 'End' });
        dataTable.addRows([
        {% for val in vals %}
        [ '{{val[0]}}' , '{{val[1]}}', {{val[2]}}, {{val[3]}} ],
         {% endfor %}
         ]);

      var options = {
      timeline: { colorByRowLabel: false },
        backgroundColor: '#ffd',
        avoidOverlappingGridLines: true
        };

      var chart = new google.visualization.Timeline(document.getElementById('gantt_div'));
      chart.draw(dataTable, options);

    }
    </script>

    {% set position = ''' + str(pos) + ''' %}
    <div id="gantt_div" style="position: absolute; left: {{position[0]}}%; top:{{position[1]}}%;
        width: ''' + str(chart_width) + '''px; height: 1000px;"></div>


    </head>
    <body>
        <h1> Savu MPI performance tests </h1>
        <div class="boxedLink">
            {% for link in links %}
            <br>Run {{loop.index}} : <a href={{link}}>here</a></br>
            {% endfor %}
        </div>
    </body>
    </html>
    '''
    return template_string


def set_template_string_vis(nVals, title, size, params, header_shift):
    import numpy as np

    total_size = list(size[0])
    total_size[1] = total_size[1]
    sub_size = list(size[1])

    plot_per_x = (1 if nVals is 1 else 2)
    nSubplots = int(np.ceil(float(nVals)/plot_per_x))*plot_per_x

    size = [total_size[0]/plot_per_x, total_size[1]/(nSubplots/plot_per_x)]
    size = [size[i]*(sub_size[i]/100.0) for i in range(len(size))]

    start = [(100 - t)/2 for t in total_size]
    gap_x = (total_size[0] - plot_per_x*size[0])/2
    gap_y = (total_size[1] - (nSubplots/plot_per_x)*size[1])/2

    right = [(start[0] + (i*2 + 1)*gap_x + i*size[0]) for i in range(plot_per_x)]*(nSubplots/plot_per_x)
    top = [(start[1] + (i*2 + 1)*gap_y + i*size[1] + header_shift) for i in range(plot_per_x)]
    top = np.reshape(np.transpose(np.tile(np.asarray(top), (nSubplots/plot_per_x, 1))), nSubplots).tolist()

    data_names = []
    chart_names = []
    for i in range(nVals):
        data_names.append('data' + str(i))
        chart_names.append('chart' + str(i))

    chartsize = [85 for s in size]

    template_string = '''
        <html>
        <head>
         <link rel="stylesheet" type="text/css" href="{{style_sheet}}">

        <title> Savu MPI performance tests </title>

        <script src="https://www.gstatic.com/charts/loader.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/rgbcolor.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/StackBlur.js"></script>
        <script type="text/javascript" src="http://canvg.googlecode.com/svn/trunk/canvg.js"></script>
        <script type="text/javascript">
        google.setOnLoadCallback(drawSeriesChart);

        function drawSeriesChart() {
            {% set title = ''' + str(title) + ''' %}

            {% for vals in frame %}
            var data{{loop.index0}} = google.visualization.arrayToDataTable([
            ['file system', '#Cores', 'Time (mean)', '#Nodes', 'Time (std)', 'link'],
              {% for val in vals %}
                [ '', {{val[1]}}, {{val[2]}}, {{val[3]}}, {{val[4]}}, '{{val[5]}}'],
              {% endfor %}
            ]);

        var view{{loop.index0}} = new google.visualization.DataView(data{{loop.index0}});
        view{{loop.index0}}.setColumns([0, 1, 2, 3, 4]);


        var options{{loop.index0}} = {
        {% set chartsize = ''' + str(chartsize) + '''%}
        chartArea: {'width': '{{chartsize[0]}}%', 'height': '{{chartsize[1]}}%'},
        colorAxis: {colors: ['yellow', 'red']},
        title: '{{title[loop.index0]}}',
        titleTextStyle: {fontSize: '15'},
        hAxis:
	  {
          title: 'Total cores',
          titleTextStyle: {fontSize: '15'},
          viewWindow: {min: 0, max: 50},
        },
        vAxis:
        {
          title: 'Mean time (seconds)',
          titleTextStyle: {fontSize: '15'},
        },
        bubble: {textStyle: {fontSize: 13}},
        sizeAxis: {maxValue: {{max_bubble}}, minValue: 0, maxSize: 40, minSize: 5}
      };

            var chart{{loop.index0}} = new google.visualization.BubbleChart(document.getElementById('chart{{loop.index0}}'));
            chart{{loop.index0}}.draw(view{{loop.index0}}, options{{loop.index0}});

         var selectHandler{{loop.index0}} = function(e) {
         window.location = data{{loop.index0}}.getValue(chart{{loop.index0}}.getSelection()[0]['row'], 5 );
        }

        google.visualization.events.addListener(chart{{loop.index0}}, 'select', selectHandler{{loop.index0}});

        {% endfor %}
        }
        </script>

        {% for vals in frame %}
        {% set right = ''' + str(right) + ''' %}
        {% set top = ''' + str(top) + ''' %}
        <div id="chart{{loop.index0}}" style="position: absolute; left: {{right[loop.index0]}}%; top:{{top[loop.index0]}}%;
        width: ''' + str(size[0]) + '''%; height: ''' + str(size[1]) + '''%;"></div>
        {% endfor %}

    </head>
    <body>
        <h1> Savu MPI performance tests </h1>
        {% set dict = ''' + str(params) + ''' %}
        {% set nDict = ''' + str(len(params)) + '''%}
        {% if nDict > 0 %}
        <div class="boxed">
            ---------
            {% for key in dict %}
            {{key}} : {{dict[key]}} ---------
            {% endfor %}
        </div>
        {% endif %}
    </body>
    </html>
    '''

    return template_string



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: temp_loader
   :platform: Unix
   :synopsis: A full-field tomography loader that creates a NeXus file in \
   NXtomo format and contains a random number generated hdf5 dataset of a \
   specified size.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.plugins.utils import register_plugin
from savu.plugins.loaders.random_hdf5_loader import RandomHdf5Loader
from savu.data.data_structures.data_types.data_plus_darks_and_flats \
    import ImageKey


@register_plugin
class Random3dTomoLoader(RandomHdf5Loader):
    """
    A hdf5 dataset of a specified size is created at runtime using numpy\
    random sampling (numpy.random), saved with relevant meta_data to a NeXus \
    file, and used as input. It recreates the behaviour of the nxtomo loader \
    but with random data.  The input file path passed to Savu will be ignored \
    (use a dummy).
    Note: Further extensions planned to allow the generated data to be \
    re-loaded with the nxtomo_loader.

    :param axis_labels: A list of axis labels. Default: ['rotation_angle.degrees', 'detector_y.angles', 'detector_x.angles'].
    :*param patterns: Patterns. Default: ['SINOGRAM.0c.1s.2c', 'PROJECTION.0s.1c.2c'].
    :*param dataset_name: The name assigned to the dataset. Default: 'tomo'.
    :u*param image_key: Specify position of darks and flats (in that order) \
    in the data. Default: [[0, 1], [2, 3]]
    """

    def __init__(self, name='Random3dTomoLoader'):
        super(Random3dTomoLoader, self).__init__(name)

    def setup(self):
        data_obj = super(Random3dTomoLoader, self).setup()
        image_key = self.__set_image_key(data_obj)
        data_obj.data = ImageKey(data_obj, image_key, 0)
        data_obj.set_shape(data_obj.data.shape)
        self.set_data_reduction_params(data_obj)
        data_obj.data._set_dark_and_flat()
        data_obj.data.update_dark(np.zeros(data_obj.data.dark().shape))
        data_obj.data.update_flat(np.ones(data_obj.data.flat().shape))

    def __set_image_key(self, data_obj):
        proj_slice = \
            data_obj.get_data_patterns()['PROJECTION']['slice_dims'][0]
        image_key = np.zeros(data_obj.data.shape[proj_slice], dtype=int)
        dark, flat = self.parameters['image_key']
        image_key[np.array(dark)] = 2
        image_key[np.array(flat)] = 1
        return image_key

    def _set_rotation_angles(self, data_obj, nEntries):
        dark, flat = self.parameters['image_key']
        nEntries = nEntries - len(dark + flat)
        super(Random3dTomoLoader, self)._set_rotation_angles(
                data_obj, nEntries)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: astra_recon_cpu
   :platform: Unix
   :synopsis: Wrapper around the Astra toolbox for cpu reconstruction
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.reconstructions.astra_recons.base_astra_recon \
    import BaseAstraRecon
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class AstraReconCpu(BaseAstraRecon, CpuPlugin):
    """
    A Plugin to run the astra reconstruction

    :u*param algorithm: Reconstruction type \
        (FBP|SIRT|SART|ART|CGLS|FP|BP|). Default: 'FBP'.
    :u*param FBP_filter: The FBP reconstruction filter type (none|ram-lak|\
        shepp-logan|cosine|hamming|hann|tukey|lanczos|triangular|gaussian|\
        barlett-hann|blackman|nuttall|blackman-harris|blackman-nuttall|\
        flat-top|kaiser|parzen). Default: 'ram-lak'.
    :param projector: Set astra projector (line|strip|linear). Default: 'line'.
    """

    def __init__(self):
        super(AstraReconCpu, self).__init__("AstraReconCpu")

    def astra_setup(self):
        options_list = ["FBP", "SIRT", "SART", "ART", "CGLS", "FP", "BP"]
        if not options_list.count(self.parameters['algorithm']):
            raise Exception("Unknown Astra CPU algorithm.")

    def set_options(self, cfg):
        return cfg



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: filter
   :platform: Unix
   :synopsis: A base class for all standard filters

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin


class BaseFilter(Plugin):
    """
    A Plugin to apply a simple dark and flatfield correction to some
    raw timeseries data
    """

    def __init__(self, name="BaseFilter"):
        super(BaseFilter, self).__init__(name)

    def get_plugin_pattern(self):
        return 'PROJECTION'

    def raw_data(self):
        """ Return True if the output dataset should retain ImageKey/NoImageKey
        instances if they exist, i.e. keep the darks and flats
        NB. This is only available if out_dataset is created from an in_dataset
        """
        return False

    def setup(self):
        self.exp.log(self.name + " Start")
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        # copy all required information from in_dataset[0]
        # raw=True/False => output is still raw data (retain darks and flats)
        out_dataset[0].create_dataset(in_dataset[0], raw=self.raw_data())

        # set information relating to the plugin data
        in_pData, out_pData = self.get_plugin_datasets()
        # set pattern_name and nframes to process for all datasets
        plugin_pattern = self.get_plugin_pattern()
        in_pData[0].plugin_data_setup(plugin_pattern, self.get_max_frames())
        out_pData[0].plugin_data_setup(plugin_pattern, self.get_max_frames())

        self.exp.log(self.name + " End")

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'multiple'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# the information below is used in the auto-API creation for documentation
# and the synopsis appears in the configurator.
"""
.. module:: plugin_template
   :platform: Unix
   :synopsis: A template to create a simple plugin that takes one dataset as\
   input and returns a similar dataset as output.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

# this decorator is required for the configurator to recognise the plugin
#@register_plugin
# Each class must inherit from the Plugin class and a driver
class PluginTemplate1(Plugin, CpuPlugin):
    """
    A description of the plugin - the synopsis above appears as the short
    plugin description in the configurator and anything written here
    appears as the longer description (-vv flag).

    :param example: Example of a plugin parameter. Default: None.

    """

    def __init__(self):
        super(PluginTemplate1, self).__init__('PluginTemplate1')

    def nInput_datasets(self):
        # Called immediately after the plugin is loaded in to the framework
        return 1

    def nOutput_datasets(self):
        # Called immediately after the plugin is loaded in to the framework
        return 1

    def setup(self):
        # This method is called after the number of in/out datasets associated
        # with the plugin has been established.  It tells the framework all
        # the information it needs to know about the data transport to-and-from
        # the plugin.


        #================== Input and output datasets =========================
        # in_datasets and out_datasets are instances of the Data class.
        # in_datasets were either created in the loader or as output from
        # previous plugins.  out_datasets objects have already been created at
        # this point, but they are empty and need to be populated.

        # Get the Data instances associated with this plugin
        in_dataset, out_dataset = self.get_datasets()

        # see https://savu.readthedocs.io/en/latest/api/savu.data.data_structures.data_create/
        # for more information on creating datasets.

        # Populate the output dataset(s)
        out_dataset[0].create_dataset(in_dataset[0])


        #================== Input and output plugin datasets ==================
        # in_pData and out_pData are instances of the PluginData class.
        # All in_datasets and out_datasets above have an in/out_pData object
        # attached to them temporarily for the duration of the plugin,
        # giving access to additional plugin-specific dataset details. At this
        # point they have been created but not yet populated.

        # Get the PluginData instances attached to the Data instances above
        in_pData, out_pData = self.get_plugin_datasets()

        # Each plugin dataset must call this method and define the data access
        # pattern and number of frames required.
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')

        # 'single', 'multiple' or an int (should only be used if essential)
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        
        # All dataset information can be accessed via the Data and PluginData
        # instances

    def pre_process(self):
        # This method is called once before any processing has begun.
        # Access parameters from the doc string in the parameters dictionary
        # e.g. self.parameters['example']
        pass

    def process_frames(self, data):
        # This function is called in a loop by the framework until all the
        # data has been processed.
    
        # Each iteration of the loop will receive a list of numpy arrays
        # (data) containing nInput_datasets with the data sliced as requested
        # in the setup method (SINOGRAM in this case).  If 'multiple' or an 
        # integer number of max_frames are requested the array with have an
        # extra dimension.

        # This plugin has one output dataset, so a single numpy array (a
        # SINOGRAM in this case) should be returned to the framework.
        return data[0]

    def post_process(self):
        # This method is called once after all processing has completed
        # (after an MPI barrier).
        pass




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_recon
   :platform: Unix
   :synopsis: A base class for all reconstruction methods

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import math
import numpy as np

import savu.core.utils as cu
from savu.plugins.plugin import Plugin

MAX_OUTER_PAD = 2.1


class BaseRecon(Plugin):
    """
    A base class for reconstruction plugins

    :u*param centre_of_rotation: Centre of rotation to use for the \
    reconstruction. Default: 0.0.

    :u*param init_vol: Dataset to use as volume initialiser \
    (doesn't currently work with preview). Default: None.

    :param centre_pad: Pad the sinogram to centre it in order to fill the \
    reconstructed volume ROI for asthetic purposes.\
    NB: Only available for selected algorithms and will be ignored otherwise. \
    WARNING: This will significantly increase the size of the data and the \
    time to compute the reconstruction). Default: False.

    :param outer_pad: Pad the sinogram width to fill the reconstructed volume \
    for asthetic purposes.\
    Choose from True (defaults to sqrt(2)), False or float <= 2.1. \
    NB: Only available for selected algorithms and will be ignored otherwise.\
    WARNING: This will increase the size of the data and the \
    time to compute the reconstruction). Default: False.

    :u*param log: Take the log of the data before reconstruction \
    (True or False). Default: True.

    :u*param preview: A slice list of required frames. Default: [].

    :param force_zero: Set any values in the reconstructed image outside of \
    this range to zero. Default: [None, None].

    :param ratio: Ratio between the diameter of a circle mask and the width of\
    a reconstructed image. If passed as a list or tuple, the second value is \
    assigned to the outer mask area, e.g [0.95, 0.0]. Default: 0.95.

    :param log_func: Override the default log \
        function. Default: 'np.nan_to_num(-np.log(sino))'.

    :param vol_shape: Override the size of the reconstruction volume with an \
    integer value. Default: 'fixed'.
    """

    def __init__(self, name='BaseRecon'):
        super(BaseRecon, self).__init__(name)
        self.nOut = 1
        self.nIn = 1
        self.scan_dim = None
        self.rep_dim = None
        self.br_vol_shape = None
        self.frame_angles = None
        self.frame_cors = None
        self.frame_init_data = None
        self.centre = None
        self.base_pad_amount = None
        self.padding_alg = False
        self.cor_shift = 0
        self.init_vol = False
        self.cor_as_dataset = False

    def base_pre_process(self):
        in_data, out_data = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        self.pad_dim = \
            in_pData[0].get_data_dimension_by_axis_label('x', contains=True)
        in_meta_data = self.get_in_meta_data()[0]
        self.__set_padding_alg()

        self.exp.log(self.name + " End")
        self.br_vol_shape = out_pData[0].get_shape()
        self.set_centre_of_rotation(in_data[0], in_meta_data, in_pData[0])

        self.main_dir = in_data[0].get_data_patterns()['SINOGRAM']['main_dir']
        self.angles = in_meta_data.get('rotation_angle')
        if len(self.angles.shape) != 1:
            self.scan_dim = in_data[0].get_data_dimension_by_axis_label('scan')
        self.slice_dirs = out_data[0].get_slice_dimensions()

        shape = in_pData[0].get_shape()
        factor = self.__get_outer_pad()
        self.sino_pad = int(math.ceil(factor * shape[self.pad_dim]))

        self.sino_func, self.cor_func = self.set_function(shape) if \
            self.padding_alg else self.set_function(False)

        self.range = self.parameters['force_zero']
        self.fix_sino = self.get_sino_centre_method()

    def __get_outer_pad(self):
        # length of diagonal of square is side*sqrt(2)
        factor = math.sqrt(2) - 1
        pad = self.parameters['outer_pad'] if 'outer_pad' in \
            list(self.parameters.keys()) else False

        if pad is not False and not self.padding_alg:
            msg = 'This reconstruction algorithm cannot be padded.'
            cu.user_message(msg)
            return 0

        if isinstance(pad, bool):
            return factor if pad is True else 0
        factor = float(pad)
        if factor > MAX_OUTER_PAD:
            factor = MAX_OUTER_PAD
            msg = 'Maximum outer_pad value is 2.1, using this instead'
            cu.user_message(msg)
        return float(pad)

    def __set_padding_alg(self):
        """ Determine if this is an algorithm that allows sinogram padding. """
        pad_algs = self.get_padding_algorithms()
        alg = self.parameters['algorithm'] if 'algorithm' in \
            list(self.parameters.keys()) else None
        self.padding_alg = True if alg in pad_algs else False

    def get_vol_shape(self):
        return self.br_vol_shape

    def set_centre_of_rotation(self, inData, mData, pData):
        # if cor has been passed as a dataset then do nothing
        if isinstance(self.parameters['centre_of_rotation'], str):
            return
        if 'centre_of_rotation' in list(mData.get_dictionary().keys()):
            cor = self.__set_cor_from_meta_data(mData, inData)
        else:
            val = self.parameters['centre_of_rotation']
            if isinstance(val, dict):
                cor = self.__polyfit_cor(val, inData)
            else:
                sdirs = inData.get_slice_dimensions()
                cor = np.ones(np.prod([inData.get_shape()[i] for i in sdirs]))
                # if centre of rotation has not been set then fix it in the
                # centre
                val = val if val != 0 else \
                    (self.get_vol_shape()[self._get_detX_dim()]) / 2.0
                cor *= val
                # mData.set('centre_of_rotation', cor) see Github ticket
        self.cor = cor
        self.centre = self.cor[0]

    def __set_cor_from_meta_data(self, mData, inData):
        cor = mData.get('centre_of_rotation')
        sdirs = inData.get_slice_dimensions()
        total_frames = np.prod([inData.get_shape()[i] for i in sdirs])
        if total_frames > len(cor):
            cor = np.tile(cor, total_frames / len(cor))
        return cor

    def __polyfit_cor(self, cor_dict, inData):
        if 'detector_y' in list(inData.meta_data.get_dictionary().keys()):
            y = inData.meta_data.get('detector_y')
        else:
            yDim = inData.get_data_dimension_by_axis_label('detector_y')
            y = np.arange(inData.get_shape()[yDim])

        z = np.polyfit(list(map(int, list(cor_dict.keys()))), list(cor_dict.values()), 1)
        p = np.poly1d(z)
        cor = p(y)
        return cor

    def set_function(self, pad_shape):
        if not pad_shape:
            def cor_func(cor): return cor
            if self.parameters['log']:
                sino_func = self.__make_lambda()
            else:
                sino_func = self.__make_lambda(log=False)
        else:
            def cor_func(cor): return cor + self.sino_pad
            if self.parameters['log']:
                sino_func = self.__make_lambda(pad=pad_shape)
            else:
                sino_func = self.__make_lambda(pad=pad_shape, log=False)
        return sino_func, cor_func

    def __make_lambda(self, log=True, pad=False):
        log_func = 'np.nan_to_num(sino)' if not log else self.parameters['log_func']
        if pad:
            pad_tuples, mode = self.__get_pad_values(pad)
            log_func = log_func.replace(
                    'sino', 'np.pad(sino, %s, "%s")' % (pad_tuples, mode))
        return eval("lambda sino: " + log_func)

    def __get_pad_values(self, pad_shape):
        mode = 'edge'
        pad_tuples = [(0, 0)] * (len(pad_shape) - 1)
        pad_tuples.insert(self.pad_dim, (self.sino_pad, self.sino_pad))
        pad_tuples = tuple(pad_tuples)
        return pad_tuples, mode

    def base_process_frames_before(self, data):
        """
        Reconstruct a single sinogram with the provided centre of rotation
        """
        sl = self.get_current_slice_list()[0]
        init = data[1] if self.init_vol else None
        angles = \
            self.angles[:, sl[self.scan_dim]] if self.scan_dim else self.angles

        self.frame_angles = angles

        dim_sl = sl[self.main_dir]

        if self.cor_as_dataset:
            self.frame_cors = self.cor_func(data[len(data) - 1])
        else:
            frame_nos = \
                self.get_plugin_in_datasets()[0].get_current_frame_idx()
            a = self.cor[tuple([frame_nos])]
            self.frame_cors = self.cor_func(a)

        # for extra padded frames that make up the numbers
        if not self.frame_cors.shape:
            self.frame_cors = np.array([self.centre])

        len_data = len(np.arange(dim_sl.start, dim_sl.stop, dim_sl.step))

        missing = [self.centre] * (len(self.frame_cors) - len_data)
        self.frame_cors = np.append(self.frame_cors, missing)

        # fix to remove NaNs in the initialised image
        if init is not None:
            init[np.isnan(init)] == 0.0
        self.frame_init_data = init

        data[0] = self.fix_sino(self.sino_func(data[0]), self.frame_cors[0])
        return data

    def base_process_frames_after(self, data):
        lower_range, upper_range = self.range
        if lower_range is not None:
            data[data < lower_range] = 0
        if upper_range is not None:
            data[data > upper_range] = 0
        return data

    def get_padding_algorithms(self):
        """ A list of algorithms that allow the data to be padded. """
        return []

    def pad_sino(self, sino, cor):
        """  Pad the sinogram so the centre of rotation is at the centre. """
        detX = self._get_detX_dim()
        pad = self.get_centre_offset(sino, cor, detX)
        self.cor_shift = pad[0]
        pad_tuples = [(0, 0)] * (len(sino.shape) - 1)
        pad_tuples.insert(detX, tuple(pad))
        self.__set_pad_amount(max(pad))
        return np.pad(sino, tuple(pad_tuples), mode='edge')

    def _get_detX_dim(self):
        pData = self.get_plugin_in_datasets()[0]
        return pData.get_data_dimension_by_axis_label('x', contains=True)

    def get_centre_offset(self, sino, cor, detX):
        centre_pad = self.br_array_pad(cor, sino.shape[detX])
        sino_width = sino.shape[detX]
        new_width = sino_width + max(centre_pad)
        sino_pad = int(math.ceil(float(sino_width) / new_width * self.sino_pad) // 2)
        pad = np.array([sino_pad]*2) + centre_pad
        return pad

    def get_centre_shift(self, sino, cor):
        detX = self._get_detX_dim()
        return max(self.get_centre_offset(sino, self.centre, detX))

    def crop_sino(self, sino, cor):
        """  Crop the sinogram so the centre of rotation is at the centre. """
        detX = self._get_detX_dim()
        start, stop = self.br_array_pad(cor, sino.shape[detX])[::-1]
        self.cor_shift = -start
        sl = [slice(None)] * len(sino.shape)
        sl[detX] = slice(start, sino.shape[detX] - stop)
        sino = sino[tuple(sl)]
        self.set_mask(sino.shape)
        return sino

    def br_array_pad(self, ctr, nPixels):
        width = nPixels - 1.0
        alen = ctr
        blen = width - ctr
        mid = (width - 1.0) / 2.0
        shift = round(abs(blen - alen))
        p_low = 0 if (ctr > mid) else shift
        p_high = shift + 0 if (ctr > mid) else 0
        return np.array([int(p_low), int(p_high)])

    def keep_sino(self, sino, cor):
        """ No change to the sinogram """
        return sino

    def get_sino_centre_method(self):
        centre_pad = self.keep_sino
        if 'centre_pad' in list(self.parameters.keys()):
            cpad = self.parameters['centre_pad']
            if not (cpad is True or cpad is False):
                raise Exception('Unknown value for "centre_pad", please choose'
                                ' True or False.')
            centre_pad = self.pad_sino if cpad and self.padding_alg \
                else self.crop_sino
        return centre_pad

    def __set_pad_amount(self, pad_amount):
        self.base_pad_amount = pad_amount

    def get_pad_amount(self):
        return self.base_pad_amount

    def get_fov_fraction(self, sino, cor):
        """ Get the fraction of the original FOV that can be reconstructed due\
        to offset centre """
        pData = self.get_plugin_in_datasets()[0]
        detX = pData.get_data_dimension_by_axis_label('x', contains=True)
        original_length = sino.shape[detX]
        shift = self.get_centre_shift(sino, cor)
        return (original_length - shift) / float(original_length)

    def get_reconstruction_alg(self):
        return None

    def get_angles(self):
        """ Get the angles associated with the current sinogram(s).

        :returns: Angles of the current frames.
        :rtype: np.ndarray
        """
        return self.frame_angles

    def get_cors(self):
        """
        Get the centre of rotations associated with the current sinogram(s).

        :returns: Centre of rotation values for the current frames.
        :rtype: np.ndarray
        """
        return self.frame_cors + self.cor_shift

    def set_mask(self, shape):
        pass

    def get_initial_data(self):
        """
        Get the initial data (if it is exists) associated with the current \
        sinogram(s).

        :returns: The section of the initialisation data associated with the \
            current frames.
        :rtype: np.ndarray or None
        """
        return self.frame_init_data

    def get_frame_params(self):
        params = [self.get_cors(), self.get_angles(), self.get_vol_shape(),
                  self.get_initial_data()]
        return params

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        # reduce the data as per data_subset parameter
        self.preview_flag = \
            self.set_preview(in_dataset[0], self.parameters['preview'])

        self._set_volume_dimensions(in_dataset[0])
        axis_labels = self._get_axis_labels(in_dataset[0])
        shape = self._get_shape(in_dataset[0])

        # output dataset
        out_dataset[0].create_dataset(axis_labels=axis_labels, shape=shape)
        out_dataset[0].add_volume_patterns(*self._get_volume_dimensions())

        # set information relating to the plugin data
        in_pData, out_pData = self.get_plugin_datasets()

        self.init_vol = 1 if 'init_vol' in list(self.parameters.keys()) and\
            self.parameters['init_vol'] else 0
        self.cor_as_dataset = 1 if isinstance(
            self.parameters['centre_of_rotation'], str) else 0
            
        for i in range(len(in_dataset) - self.init_vol - self.cor_as_dataset):
            in_pData[i].plugin_data_setup('SINOGRAM', self.get_max_frames(),
                                          slice_axis=self.get_slice_axis())
            idx = 1

        # initial volume dataset
        if self.init_vol:
#            from savu.data.data_structures.data_types import Replicate
#            if self.rep_dim:
#                in_dataset[idx].data = Replicate(
#                    in_dataset[idx], in_dataset[0].get_shape(self.rep_dim))
            in_pData[1].plugin_data_setup('VOLUME_XZ', self.get_max_frames())
            idx += 1

        # cor dataset
        if self.cor_as_dataset:
            self.cor_as_dataset = True
            in_pData[idx].plugin_data_setup('METADATA', self.get_max_frames())

        # set pattern_name and nframes to process for all datasets
        out_pData[0].plugin_data_setup('VOLUME_XZ', self.get_max_frames())

    def _get_axis_labels(self, in_dataset):
        """
        Get the new axis labels for the output dataset - this is now a volume.

        Parameters
        ----------
        in_dataset : :class:`savu.data.data_structures.data.Data`
            The input dataset to the plugin.

        Returns
        -------
        labels : dict
            The axis labels for the dataset that is output from the plugin.

        """
        labels = in_dataset.data_info.get('axis_labels')[0]
        volX, volY, volZ = self._get_volume_dimensions()
        labels = [str(volX) + '.voxel_x.voxels', str(volZ) + '.voxel_z.voxels']
        if volY:
            labels.append(str(volY) + '.voxel_y.voxels')
        labels = {in_dataset: labels}
        return labels

    def _set_volume_dimensions(self, data):
        """
        Map the input dimensions to the output volume dimensions

        Parameters
        ----------
        in_dataset : :class:`savu.data.data_structures.data.Data`
            The input dataset to the plugin.
        """
        data._finalise_patterns()
        self.volX = data.get_data_dimension_by_axis_label("rotation_angle")
        self.volZ = data.get_data_dimension_by_axis_label("x", contains=True)
        self.volY = data.get_data_dimension_by_axis_label(
            "y", contains=True, exists=True)

    def _get_volume_dimensions(self):
        return self.volX, self.volY, self.volZ

    def _get_shape(self, in_dataset):
        shape = list(in_dataset.get_shape())
        volX, volY, volZ = self._get_volume_dimensions()

        if self.parameters['vol_shape'] in ('auto', 'fixed'):
            shape[volX] = shape[volZ]
        else:
            shape[volX] = self.parameters['vol_shape']
            shape[volZ] = self.parameters['vol_shape']

        if 'resolution' in self.parameters.keys():
            shape[volX] /= self.parameters['resolution']
            shape[volZ] /= self.parameters['resolution']
        return tuple(shape)

    def get_max_frames(self):
        """
        Number of data frames to pass to each instance of process_frames func

        Returns
        -------
        str or int
            "single", "multiple" or integer (only to be used if the number of
                                             frames MUST be fixed.)
        """
        return 'multiple'
    
    def get_slice_axis(self):
        """
        Fix the fastest changing slice dimension

        Returns
        -------
        str or None
            str should be the axis_label corresponding to the fastest changing
            dimension

        """
        return None

    def nInput_datasets(self):
        nIn = 1
        if 'init_vol' in self.parameters.keys() and \
                self.parameters['init_vol']:
            if len(self.parameters['init_vol'].split('.')) == 3:
                name, temp, self.rep_dim = self.parameters['init_vol']
                self.parameters['init_vol'] = name
            nIn += 1
            self.parameters['in_datasets'].append(self.parameters['init_vol'])
        if isinstance(self.parameters['centre_of_rotation'], str):
            self.parameters['in_datasets'].append(
                self.parameters['centre_of_rotation'])
            nIn += 1
        return nIn

    def nOutput_datasets(self):
        return self.nOut

    def reconstruct_pre_process(self):
        """
        Should be overridden to perform pre-processing in a child class
        """
        pass

    def executive_summary(self):
        summary = []
        if not self.preview_flag:
            summary.append(("WARNING: Ignoring preview parameters as a preview"
                            " has already been applied to the data."))
        if len(summary) > 0:
            return summary
        return ["Nothing to Report"]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: paganin_filter
   :platform: Unix
   :synopsis: A plugin to apply the Paganin filter.

.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
import math
import logging
import numpy as np
import pyfftw.interfaces.scipy_fftpack as fft

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin, dawn_compatible
from savu.data.plugin_list import CitationInformation


@dawn_compatible
@register_plugin
class PaganinFilter(BaseFilter, CpuPlugin):
    """
    A plugin to apply Paganin filter (contrast enhancement) on projections.

    :param Energy: Given X-ray energy in keV. Default: 53.0.
    :param Distance: Distance from sample to detection - Unit is \
        metre. Default: 1.0.
    :param Resolution: Pixel size - Unit is micron. Default: 1.28.
    :u*param Ratio: ratio of delta/beta. Default: 250.0.
    :param Padtopbottom: Pad to the top and bottom of projection. Default: 10.
    :param Padleftright: Pad to the left and right of projection. Default: 10.
    :param Padmethod: Numpy pad method. Default: 'edge'.
    :param increment: Increment all values by this amount before taking the \
        log. Default: 0.0.

    :config_warn: The 'log' parameter in the reconstruction should be set to \
    FALSE.
    :config_warn: Previewing a subset of sinograms will alter the result, due \
    to the global nature of this filter. If this is necessary, ensure they \
    are consecutive.

    """

    def __init__(self):
        logging.debug("initialising Paganin Filter")
        logging.debug("Calling super to make sure that all superclases are " +
                      " initialised")
        super(PaganinFilter, self).__init__("PaganinFilter")
        self.filtercomplex = None
        self.count = 0

    def set_filter_padding(self, in_pData, out_pData):
        in_data = self.get_in_datasets()[0]
        det_x = in_data.get_data_dimension_by_axis_label('detector_x')
        det_y = in_data.get_data_dimension_by_axis_label('detector_y')
        pad_det_y = '%s.%s' % (det_y, self.parameters['Padtopbottom'])
        pad_det_x = '%s.%s' % (det_x, self.parameters['Padleftright'])
        mode = self.parameters['Padmethod']
        pad_dict = {'pad_directions': [pad_det_x, pad_det_y], 'pad_mode': mode}
        in_pData[0].padding = pad_dict
        out_pData[0].padding = pad_dict

    def pre_process(self):
        self._setup_paganin(*self.get_plugin_in_datasets()[0].get_shape())

    def _setup_paganin(self, height, width):
        micron = 10**(-6)
        keV = 1000.0
        distance = self.parameters['Distance']
        energy = self.parameters['Energy'] * keV
        resolution = self.parameters['Resolution'] * micron
        wavelength = (1240.0 / energy) * 10.0**(-9)
        ratio = self.parameters['Ratio']

        height1 = height + 2 * self.parameters['Padtopbottom']
        width1 = width + 2 * self.parameters['Padleftright']
        centery = np.ceil(height1 / 2.0) - 1.0
        centerx = np.ceil(width1 / 2.0) - 1.0

        # Define the paganin filter
        dpx = 1.0 / (width1 * resolution)
        dpy = 1.0 / (height1 * resolution)
        pxlist = (np.arange(width1) - centerx) * dpx
        pylist = (np.arange(height1) - centery) * dpy
        pxx = np.zeros((height1, width1), dtype=np.float32)
        pxx[:, 0:width1] = pxlist
        pyy = np.zeros((height1, width1), dtype=np.float32)
        pyy[0:height1, :] = np.reshape(pylist, (height1, 1))
        pd = (pxx * pxx + pyy * pyy) * wavelength * distance * math.pi

        filter1 = 1.0 + ratio * pd
        self.filtercomplex = filter1 + filter1 * 1j

    def _paganin(self, data):
        pci1 = fft.fft2(np.float32(data))
        pci2 = fft.fftshift(pci1) / self.filtercomplex
        fpci = np.abs(fft.ifft2(pci2))
        result = -0.5 * self.parameters['Ratio'] * np.log(
            fpci + self.parameters['increment'])
        return result

    def process_frames(self, data):
        proj = np.nan_to_num(data[0])  # Noted performance
        proj[proj == 0] = 1.0
        return self._paganin(proj)

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The contrast enhancement used in this processing chain is taken\
             from this work.")
        cite_info.bibtex = \
            ("@article{paganin2002simultaneous,\n" +
             "title={Simultaneous phase and amplitude extraction from a single\
             defocused image of a homogeneous object},\n" +
             "author={Paganin, David and Mayo, SC and Gureyev, Tim E and \
             Miller, Peter R and Wilkins, Steve W},\n" +
             "journal={Journal of microscopy},\n" +
             "volume={206},\n" +
             "number={1},\n" +
             "pages={33--40},\n" +
             "year={2002},\n" +
             "publisher={Wiley Online Library}" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Simultaneous phase and amplitude extraction from a single \
             defocused image of a homogeneous object\n" +
             "%A Paganin, David\n" +
             "%A Mayo, SC\n" +
             "%A Gureyev, Tim E\n" +
             "%A Miller, Peter R\n" +
             "%A Wilkins, Steve W\n" +
             "%J Journal of microscopy\n" +
             "%V 206\n" +
             "%N 1\n" +
             "%P 33-40\n" +
             "%@ 1365-2818\n" +
             "%D 2002\n" +
             "%I Wiley Online Library\n")
        cite_info.doi = "doi: DOI: 10.1046/j.1365-2818.2002.01010.x"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# $Id: dezing_filter.py 467 2016-02-16 11:40:42Z kny48981 $


"""
.. module:: dezinger_simple
   :platform: Unix
   :synopsis: A plugin to remove zingers

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

import scipy.signal.signaltools as sig

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class DezingerSimpleDeprecated(BaseFilter, CpuPlugin):
    """
    A plugin for cleaning x-ray strikes based on statistical evaluation of \
    the near neighbourhood
    :param outlier_mu: Threshold for detecting outliers, greater is less \
    sensitive. Default: 1000.0.
    :param kernel_size: Number of frames included in average - if the number \
        is not odd use kernel_size+1. Default: 5.
    """

    def __init__(self):
        super(DezingerSimpleDeprecated, self).__init__("DezingerSimpleDeprecated")
        self.zinger_proportion = 0.0
        self.frame_limit = 8

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        self.proj_dim = inData.data.proj_dim

        self._kernel = [1]*3
        self._kernel[self.proj_dim] = self.kernel_size

        pad_list = [(0, 0)]*3
        pad_list[self.proj_dim] = (self.pad, self.pad)

        dark = inData.data.dark()
        flat = inData.data.flat()
        if dark.size:
            dark = np.pad(inData.data.dark(), pad_list, mode='edge')
            dark = self._process_calibration_frames(dark)
            inData.data.update_dark(dark[self.pad:-self.pad])
        if flat.size:
            flat = np.pad(inData.data.flat(), pad_list, mode='edge')
            flat = self._process_calibration_frames(flat)
            inData.data.update_flat(flat[self.pad:-self.pad])

    def _process_calibration_frames(self, data):
        nSlices = data.shape[self.proj_dim] - 2*self.pad
        nSublists = int(np.ceil(nSlices/float(self.frame_limit)))
        idx = np.array_split(np.arange(self.pad, nSlices+self.pad), nSublists)
        idx = [np.arange(a[0]-self.pad, a[-1]+self.pad+1) for a in idx]
        out_sl = np.tile([slice(None)]*3, [len(idx), 1])
        out_sl[:, self.proj_dim] = idx
        result = np.empty_like(data)
        for sl in out_sl:
            result[tuple(sl)] = self._dezing(data[tuple(sl)])
        return result

    def _dezing(self, data):
        result = data[...]
        median_result = sig.medfilt(data, self._kernel)
        differrence = np.abs(data-median_result)
        replace_mask = differrence > self.parameters['outlier_mu']
        self.zinger_proportion = \
            max(self.zinger_proportion, np.sum(replace_mask)/(
                np.size(replace_mask)*1.0))
        result[replace_mask] = median_result[replace_mask]
        return result

    def process_frames(self, data):
        return self._dezing(data[0])

    def get_max_frames(self):
        """ Setting nFrames to multiple with an upper limit of 4 frames. """
        return ['multiple', self.frame_limit]

    def raw_data(self):
        return True

    def set_filter_padding(self, in_data, out_data):
        # kernel size must be odd
        ksize = self.parameters['kernel_size']
        self.kernel_size = ksize+1 if ksize % 2 == 0 else ksize

        in_data = in_data[0]
        self.pad = (self.kernel_size - 1) // 2
        self.data_size = in_data.get_shape()
        in_data.padding = {'pad_multi_frames': self.pad}
        out_data[0].padding = {'pad_multi_frames': self.pad}

    def executive_summary(self):
        if self.zinger_proportion > 0.5:
            return ["Over 50% of the pixels were treated as zingers!!"]
        if self.zinger_proportion > 0.05:
            return ["Over 5% of the pixels were treated as zingers"]
        return ["Nothing to Report"]



from __future__ import print_function, division

import argparse
import itertools
import os
import re
import tempfile

import numpy as np
import pandas as pd


def reset_first_entry(memory_data):
    """
    Resets the first memory entry for each unique component in the series
    """

    def _reset(s):
        try:
            for i in range(64):
                memory_data.Memory[memory_data.Key.str.contains(s.format(i))].iloc[0][1] = '0'
        except IndexError:
            pass

    _reset("CPU{}")
    _reset("GPU{}")

    return memory_data


def get_memory_data(frame):
    memory_data = frame[frame.Message.str.contains("memory usage")]
    number_finder = re.compile('[0-9]+')
    memory_data.insert(4, "Memory", [list(itertools.chain(msg.split()[0:1], number_finder.findall(msg))) for msg in
                                     memory_data.Message])

    memory_data = pd.concat([memory_data.Key, memory_data.Memory], axis=1, join="outer")
    return reset_first_entry(memory_data)


def convert(log_file_list, path, log_level):
    for log_file in log_file_list:
        the_key = ""
        the_interval = 0  # millisecs
        frame = get_frame(log_file, the_key, 'DEBUG')
        machine_names = get_machine_names(frame)
        memory_data = get_memory_data(frame)
        if log_level != 'DEBUG':
            frame = get_frame(log_file, the_key, log_level)
        html_filename = \
            set_file_name('/'.join([path, os.path.basename(log_file)]))
        render_template(frame, machine_names, memory_data, the_interval, html_filename)
        print("html file created:", html_filename)
        print("Open the html file in your browser to view the profile.")

    return frame


def get_frame(log_file, the_key, log_level):
    import itertools

    names = ['L', 'Time', 'Machine', 'CPU', 'Type', 'Message']
    data = pd.io.parsers.read_fwf(log_file, widths=[2, 13, 6, 6, 7, 1000],
                                  names=names)

    data['Key'] = data['Machine'] + data['CPU']
    frame = ((data[data.Type == log_level])[data.columns[[6, 5, 1]]])
    frame.insert(0, 'Index', list(range(len(frame))))
    frame = frame.sort_values(by=['Key', 'Index'])
    del frame['Index']

    frame.Time = frame.Time.astype(np.int32)
    startTime = (frame.groupby('Key').first()).Time
    nElems = frame.groupby('Key').size()

    shift = []
    for i in range(len(nElems)):
        shift.append([startTime[i]] * nElems[i])
    shift = list(itertools.chain(*shift))

    frame.Time = frame.Time - shift
    frame = frame[frame.Key.str.contains(the_key)]
    frame.insert(3, "Time_end", frame.Time.shift(-1))
    frame.Message = frame.Message.str.strip('\n')
    frame.Message = frame.Message.str.replace("'", "")

    return frame


def get_machine_names(frame):
    machine_names = frame.copy(deep=True)
    machine_names = machine_names[frame.Message.str.contains('Rank').replace(np.nan, False)]
    machine_names.Message = [m.split(':')[-1].strip() for m in frame.Message if isinstance(m, str) and 'Rank' in m]
    machine_names = machine_names.drop(['Time', 'Time_end'], axis=1)
    machine_names.Key = [k.split('CPU')[0] for k in machine_names.Key]
    machine_names.Key = [k.split('GPU')[0] for k in machine_names.Key]
    machine_names = machine_names.groupby('Key').first()
    machine_names['Machine'] = machine_names.index.values

    return machine_names


def render_template(frame, machine_names, memory_data, the_interval, outfilename):
    from jinja2 import Template

    frame = frame[(frame.Time_end - frame.Time) > the_interval].values

    f_out = open(outfilename, 'w')
    dirname = os.path.dirname(__file__)
    style = os.path.join(dirname, 'style_sheet.css')
    with open(os.path.join(dirname, 'string_single.html'), 'r') as template_file:
        template = Template(template_file.read())

    f_out.write(template.render(chart_width=1300, position=[16, 9],
                                vals=map(list, frame[:, 0:4]),
                                machines=map(list, machine_names.values),
                                memory=map(list, memory_data.values),
                                style_sheet=style))
    f_out.close()

    return frame


def set_file_name(filename):
    dir_path = os.path.dirname(filename)
    temp = os.path.basename(filename).split('.')
    filename = temp[0] + '_' + temp[1] + '.html'
    outfilename = dir_path + '/' + filename

    return outfilename


def __option_parser():
    """ Option parser for command line arguments.
    """
    parser = argparse.ArgumentParser(prog='savu_profile')

    parser.add_argument('file', help='Savu output log file')
    parser.add_argument('-l', '--loglevel', default='INFO',
                        help='Set the log level.')
    parser.add_argument('-f', '--find', nargs='*', default=[],
                        help='Find lines containing these entries')
    parser.add_argument('-i', '--ignore', nargs='*', default=[],
                        help='Ignore lines containing these entries')
    return parser.parse_args()


def main():
    args = __option_parser()

    filename = os.path.abspath(args.file)

    # create the log file for profiling
    name, ext = os.path.splitext(os.path.basename(filename))
    log_filename = os.path.join(tempfile.mkdtemp(), '{}_{}.log'.format(name, ext))

    lfilter = ['L '] + args.find
    with open(filename, 'r') as finput:
        with open(log_filename, 'w') as foutput:
            for line in finput:
                filter_line = [True if t in line else False for t in lfilter]
                keep_line = [False if t in line else True for t in args.ignore]
                line = False if False in filter_line + keep_line else line
                if line:
                    foutput.write(line)

    convert([log_filename], os.path.dirname(filename), args.loglevel)


if __name__ == "__main__":
    main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_type
   :platform: Unix
   :synopsis: A base module for different data types.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import inspect

import savu.plugins.utils as pu


class BaseType(object):

    def __init__(self):
        self._clone_args = {}
        self._map_args = {}
        self.base_data_args()

    def __getitem__(self, index):
        """ Override __getitem__ and map to the relevant files """
        raise NotImplementedError("__getitem__ must be implemented.")

    def get_shape(self):
        """ Get full stitched shape of a stack of files"""
        raise NotImplementedError("get_shape must be implemented.")

    def add_base_class_with_instance(self, base, inst):
        """ Add a base class instance to a class (merging of two data types).

        :params class base: a class to add as a base class
        :params instance inst: an instance of the base class
        """
        cls = self.__class__
        namespace = self.__class__.__dict__.copy()
        self.__dict__.update(inst.__dict__)
        self.__class__ = cls.__class__(cls.__name__, (cls, base), namespace)

    def base_data_args(self):
        """ Create a dictionary of required input arguments, required for
        checkpointing. """
        cls = self.__class__.__module__ + '.' + self.__class__.__name__
        args, kwargs, extras = self.clone_data_args([], {}, [])
        self._clone_args = self._set_parameters(args, kwargs, cls, cls, extras)

        args, kwargs, cls_path, extras = self.map_input_args([], {}, cls, [])
        self._map_args = self._set_parameters(
                args, kwargs, cls, cls_path, extras)

    def _set_parameters(self, args, kwargs, cls, cls_path, extras):
        mod = '.'.join(cls_path.split('.')[:-1])
        cls = cls_path.split('.')[-1]
        func = pu.load_class(mod, cls).__init__

        argspec = inspect.getfullargspec(func)
        if len(argspec[0])-1 != len(args) + len(list(kwargs.keys())):
            raise Exception('Incorrect number of input arguments mapped.')

        data_lists = [True if isinstance(a, list) and isinstance(a[0], int)
                      else False for a in args]

        # If there are multiple data objects this is incompatible with
        # checkpointing.
        ddict = {}
        if args.count('self') > 2 or data_lists.count(True):
            ddict = None
        else:
            ddict['args'] = args
            ddict['kwargs'] = kwargs
            ddict['cls'] = cls_path
            ddict['extras'] = extras
        return ddict

    def clone_data_args(self, args, kwargs, extras):
        """ Gather all information required to keep this dataset after a
        plugin has completed (may require a conversion to a different
        data_type.
        """
        raise NotImplementedError("clone_data_args must be implemented.")

    def map_input_args(self, args, kwargs, cls, extras):
        """ Gather all information required to recreate a datatype: For
        checkpointing and cloning """
        args, kwargs, extras = self.clone_data_args(args, kwargs, extras)
        return args, kwargs, cls, extras

    def get_clone_args(self):
        return self._clone_args

    def get_map_args(self):
        return self._map_args

    def create_next_instance(self, newObj):
        dtype_dict = self.get_map_args()
        if dtype_dict is None:
            return None
        args, kwargs, cls, extras = self._get_parameters(dtype_dict)
        args = [newObj if isinstance(a, str) and a == 'self'
                else a for a in args]

        cls_split = cls.split('.')
        mod = '.'.join(cls_split[:-1])
        name = cls_split[-1]
        cls_inst = pu.load_class(mod, cls_name=name)
        newObj.data = cls_inst(*args, **kwargs)
        self._base_post_clone_updates(newObj.data, extras)

    def _base_post_clone_updates(self, obj, extras):
        self.__update_extra_params(obj, extras)
        self._post_clone_updates()

    def _post_clone_updates(self):
        pass

    def _get_parameters(self, dtype_dict):
        args = dtype_dict['args']
        kwargs = dtype_dict['kwargs']
        extras = dtype_dict['extras']
        args = [self._str_to_value(self, a) for a in args]

        extras = self.__get_extras_vals(extras)
        for key, value in kwargs.items():
            kwargs[key] = self._str_to_value(self, value)
        return args, kwargs, dtype_dict['cls'], extras

    def _str_to_value(self, obj, val):
        if not isinstance(val, str):
            return val
        if val == 'self':
            return val
        val = 'self' if val == 'DATA_OBJECT' else val
        val = val.replace('self.', '')
        temp = getattr(obj, val, val)
        return temp() if inspect.ismethod(temp) else temp

    def __get_extras_vals(self, vals):
        extras = {}
        for key in vals:
            extras[key] = getattr(self, key)
        return extras

    def __update_extra_params(self, newObj, extras):
        for key, value in extras.items():
            setattr(newObj, key, value)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ica
   :platform: Unix
   :synopsis: A plugin to fit peaks

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
from savu.plugins.utils import register_plugin
from savu.plugins.component_analysis.base_component_analysis \
    import BaseComponentAnalysis
from sklearn.decomposition import PCA
import numpy as np


@register_plugin
class Pca(BaseComponentAnalysis):
    """
    This plugin performs independent component analysis on XRD/XRF spectra.
    """

    def __init__(self):
        super(Pca, self).__init__("Pca")

    def process_frames(self, data):
        logging.debug("Starting the PCA")
        data = data[0]
        sh = data.shape
        newshape = (np.prod(sh[:-1]), sh[-1])
        data = np.reshape(data, (newshape))
        # data will already be shaped correctly
        logging.debug("Making the matrix")
        pca = PCA(n_components=self.parameters['number_of_components'],
                      whiten=self.parameters['whiten'])
        logging.debug("Performing the fit")
        data = self.remove_nan_inf(data)  #otherwise the fit flags up an error for obvious reasons
#         print "I'm here"
        S_ = pca.fit_transform(data)

#         print "S_Shape is:"+str(S_.shape)
#         print "self.images_shape:"+str(self.images_shape)
        loading = np.reshape(S_, (self.images_shape))
        scores = pca.components_
        logging.debug("mange-tout")
        return [loading, scores]



import os

import numpy as np
import pandas as pd

import GraphicalThreadProfiler as gtp


def convert(filename):
    the_key = ""  # CPU0"
    the_interval = 0.8  # millisecs

    all_frames = []
    for files in filename:
        frame = gtp.get_frame(files, the_key, 'DEBUG')
        [index, nth] = get_index(frame)
        frame.Time_end[frame.index[np.cumsum(nth) - 1]] = frame.Time.max()
        frame = frame.reset_index(drop=True)
        temp = pd.DataFrame(np.array(frame.Time_end - frame.Time),
                            columns=['Time_diff'])
        temp['Key'] = frame.Key
        temp['index'] = [item for sublist in index for item in sublist]
        temp = temp.pivot(index='index', columns='Key', values='Time_diff')

        all_frames.append(temp)

    avg_duration = get_average_duration(all_frames)
    stats = output_stats(all_frames, filename)

    frame.Time_end = avg_duration
    frame.Time = frame.Time_end.shift(1)
    frame.Time.iloc[0] = 0

    render_template(frame, the_interval, set_file_name(filename),
                    get_links(filename), stats)
    return frame


def get_index(frame):
    nth = []
    index = []
    for i in range(np.size(frame.Key.unique())):
        nth.append(frame[frame.Key == frame.Key.unique()[i]].count()[0])
        index.append(list(range(nth[i])))

    return [index, nth]


def get_average_duration(frames):
    avg_frames = pd.Panel({n: df for n, df in enumerate(frames)})
    avg_frames = (avg_frames.mean(axis=0)).cumsum(0)
    avg_frames = avg_frames.as_matrix(columns=avg_frames.columns[:])
    avg_frames = np.reshape(np.transpose(avg_frames),
                            np.prod(avg_frames.shape))
    avg_frames = avg_frames[~np.isnan(avg_frames)]

    return avg_frames


def set_file_name(filename):
    [dir_path, name] = get_base_path(filename)

    name = name[0].split('.')[0] + '_avg' + repr(filename) + '.html'
    outfilename = dir_path + name

    return outfilename


def get_links(filename):
    [dir_path, fname] = get_base_path(filename)

    name = [gtp.set_file_name(file) for file in fname]
    links = [(dir_path + n) for n in name]

    return links


def render_template(frame, the_interval, outfilename, theLinks, theStats):
    from jinja2 import Template

    frame = frame[(frame.Time_end - frame.Time) > the_interval].values

    f_out = open(outfilename, 'w')
    dirname = os.path.dirname(__file__)
    style = os.path.join(dirname, 'style_sheet.css')
    with open(os.path.join(dirname, 'string_single.html'), 'r') as template_file:
        template = Template(template_file.read())
    f_out.write(template.render(chart_width=1300, position=[16, 9],
                                vals=map(list, frame[:, 0:4]), links=theLinks,
                                stats=theStats, style_sheet=style))
    f_out.close()

    return frame


def output_stats(frames, filename):
    temp = []
    for frame in frames:
        temp.append(frame.sum())

    total = pd.concat([i for i in temp], axis=0)
    total_stats = total.describe()

    [dir_path, name] = get_base_path(filename)
    name = name[0].split('.')[0] + '_avg' + repr(len(filename)) + '_stats.csv'
    outfilename = dir_path + name

    total_stats.to_csv(outfilename)

    return total_stats


def get_machine_names(all_frames):
    pass


def get_base_path(filename):
    if len(filename[0].split('/')) is 0:
        dir_path = os.getcwd()
    else:
        dir_path = os.path.dirname(filename[0])
        filename = [(os.path.basename(f)) for f in filename]

    return [dir_path + '/', filename]


if __name__ == "__main__":
    import optparse

    usage = "%prog [options] input_file"
    parser = optparse.OptionParser(usage=usage)

    (options, args) = parser.parse_args()
    print(args)
    filename = [(os.getcwd() + '/' + file) for file in args]
    convert(filename)



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 .. module:: perform elemetary arithmetic operations on data: addition, subtraction, multiplication and division
   :platform: Unix
   :synopsis: perform elemetary arithmetic operations on data: addition, subtraction, multiplication and division

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""


from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

@register_plugin
class ArithmeticOperations(Plugin, CpuPlugin):
    """
    Basic arithmetic operations on data: addition, subtraction, multiplication and division.\
    Operations can be performed by extracting scalars from METADATA (min, max, mean) OR providing a scalar value.
    
    :param scalar_value: A scalar value value for arithmetic operation (it not in metadata). Default: None.
    :param operation: arithmetic operation to apply to data, choose from addition, subtraction, multiplication and division. Default: 'division'.
    :param metadata_value: A type of scalar extracted from metadata (min, max, mean). Default: 'max'.
    """

    def __init__(self):
        super(ArithmeticOperations, self).__init__("ArithmeticOperations")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')
        out_pData[0].plugin_data_setup('VOLUME_XZ', 'single')

    def pre_process(self):
        data = self.get_in_datasets()[0]
        self.scalar_res = 0.0
        try:
            the_max = data.meta_data.get(['stats', 'max', 'pattern'])
        except KeyError:
            the_max = self.parameters['scalar_value']
        try:
            the_min = data.meta_data.get(['stats', 'min', 'pattern'])
        except KeyError:
            the_min = self.parameters['scalar_value']
        try:
            the_mean = data.meta_data.get(['stats', 'mean', 'pattern'])
        except KeyError:
            the_mean = self.parameters['scalar_value']
        # working with METADATA
        if (self.parameters['metadata_value'] == 'min'):
            if (the_min is not None):
                self.scalar_res = the_min
        if (self.parameters['metadata_value'] == 'max'):
            if (the_max is not None):
                self.scalar_res = the_max
        if (self.parameters['metadata_value'] == 'mean'):
            if (the_mean is not None):
                self.scalar_res = the_mean
        
    def process_frames(self, data):
        if (self.scalar_res != 0.0):
            if (self.parameters['operation'] == 'addition'):
                corr_data = data[0] + self.scalar_res
            if (self.parameters['operation'] == 'subtraction'):
                corr_data = data[0] - self.scalar_res
            if (self.parameters['operation'] == 'multiplication'):
                corr_data = np.multiply(data[0], self.scalar_res)
            if (self.parameters['operation'] == 'division'):
                corr_data = np.true_divide(data[0], self.scalar_res)
        return corr_data
    
    def nInput_datasets(self):
        return 1
    
    def nOutput_datasets(self):
        return 1
    
    def get_max_frames(self):
        return 'single'



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: transport_setup
   :platform: Unix
   :synopsis: Module containing classes required for transport setup, e.g MPI.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import logging
import socket
import os

from mpi4py import MPI
import savu.core.utils as cu


class MPI_setup(object):

    def __init__(self, options, name='MPI_setup'):
        self.__set_dictionary(options)

    def __set_dictionary(self, options):
        """ Fill the options dictionary with MPI related values.
        """
        processes = options["process_names"].split(',')

        if len(processes) is 1:
            options["mpi"] = False
            options["process"] = 0
            options["processes"] = processes
            self.__set_logger_single(options)
        else:
            options["mpi"] = True
            self.__mpi_setup(options)

        logging.debug(options)

    def __mpi_setup(self, options):
        """ Set MPI process specific values and logging initialisation.
        """
        hosts = MPI.COMM_WORLD.allgather(socket.gethostname())
        uniq_hosts = set(hosts)
        names = options['process_names'].split(',')

        n_cores = len(hosts)
        n_nodes = len(uniq_hosts)
        n_cores_per_node = len(names)

        rank_map = [i for s in uniq_hosts for i in range(n_cores)
                    if s == hosts[i]]
        index = sorted(list(range(len(rank_map))), key=lambda k: rank_map[k])
        all_processes = [(names*n_nodes)[index[i]] for i in range(n_cores)]
        options['processes'] = all_processes
        rank = MPI.COMM_WORLD.rank
        options['process'] = rank
        node_number = rank_map.index(rank) // n_cores_per_node
        local_name = all_processes[rank]

        self.__set_logger_parallel("%03i" % node_number, local_name, options)

        MPI.COMM_WORLD.barrier()
        logging.debug("Rank : %i - Size : %i - host : %s", rank, n_cores,
                      hosts[MPI.COMM_WORLD.rank])

        IP = socket.gethostbyname(socket.gethostname())
        logging.debug("ip address is : %s", IP)
        self.call_mpi__barrier()
        logging.debug("LD_LIBRARY_PATH is %s",  os.getenv('LD_LIBRARY_PATH'))
        self.call_mpi__barrier()

    def call_mpi__barrier(self):
        """ Call MPI_barrier before an experiment is created.
        """
        logging.debug("Waiting at the _barrier")
        MPI.COMM_WORLD.barrier()

    def __set_logger_single(self, options):
        """ Set single-threaded logging.
        """
        log_format = 'L %(relativeCreated)12d M000 CPU00' + \
                     ' %(levelname)-6s %(message)s'
        filename = os.path.join(options['log_path'], 'log.txt')
        level = cu._get_log_level(options)
        self.__set_logger(level, log_format, fname=filename)
        cu.add_user_log_level()
        self.__add_user_logging(options)
        self.__add_console_logging()

    def __set_logger_parallel(self, number, rank, options):
        """ Set parallel logger.
        """
        machine = 'M%-5s%-6s' % (number, rank)
        log_format = 'L %(relativeCreated)12d ' + machine +\
                     ' %(levelname)-6s %(message)s'
        level = cu._get_log_level(options)
        self.__set_logger(level, log_format)

        # Only add user logging to the 0 rank process
        cu.add_user_log_level()
        if MPI.COMM_WORLD.rank == 0:
            self.__add_user_logging(options)
            if not options['cluster']:
                self.__add_console_logging()

    def __set_logger(self, level, fmat, fname=None):
        datefmt = '%H:%M:%S'
        if fname:
            logging.basicConfig(level=level, format=fmat, datefmt=datefmt,
                                filename=fname, filemode='w')
        else:
            logging.basicConfig(level=level, format=fmat, datefmt=datefmt)

    def __add_console_logging(self):
        console = logging.StreamHandler()
        console.setLevel(cu.USER_LOG_LEVEL)
#        formatter = \
#            logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
        formatter = \
            logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)

    def __add_user_logging(self, options):
        logger = logging.getLogger()
        filename = os.path.join(options['out_path'], 'user.log')
        cu.add_user_log_handler(logger, filename)
        if 'syslog_server' in list(options.keys()):
            try:
                cu.add_syslog_log_handler(logger, options['syslog_server'],
                                          options['syslog_port'])
            except:
                logger.warn("Unable to add syslog logging for server %s on"
                            " port %i", options['syslog_server'],
                            options['syslog_port'])



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ImageInterpolation
   :platform: Unix
   :synopsis: A plugin to interpolate each frame

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
import numpy as np
# TODO replace imresize with PIL Image
# from scipy.misc import imresize
from PIL import Image

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin

from savu.plugins.utils import register_plugin, dawn_compatible


def imresize(data, size, interp, mode):
    """
    Provides an imresize implementation, as it is removed from scipy 1.3.0
    Uses PIL.Image, following docs advice from
    https://docs.scipy.org/doc/scipy-1.2.1/reference/generated/scipy.misc.imresize.html
    """
    return np.array(Image.fromarray(data, mode).resize(size, interp))


@register_plugin
@dawn_compatible
class ImageInterpolation(BaseFilter, CpuPlugin):
    """
    A plugin to interpolate an image by a factor
    a wrapper on scipy.misc.imresize.
    :param size: int, float or tuple. Default: 2.0.
    :param interp: nearest lanczos bilinear bicubic cubic. Default:'bicubic'.
    """

    def __init__(self):
        logging.debug("Starting ImageInterpolation")
        super(ImageInterpolation,
              self).__init__("ImageInterpolation")

    def process_frames(self, data):
        data = data[0]
        return imresize(data, self.parameters['size'], self.parameters['interp'], mode=None)

    def setup(self):
        # get all in and out datasets required by the plugin
        in_dataset, out_dataset = self.get_datasets()
        # get plugin specific instances of these datasets
        in_pData, out_pData = self.get_plugin_datasets()

        in_pData[0].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())
        inshape = in_dataset[0].get_shape()
        imshape = inshape[-2:]
        restshape = inshape[:-2]
        outshape = imresize(np.ones(imshape), self.parameters['size'], self.parameters['interp'], None).shape
        outshape = restshape + outshape
        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=outshape)

        out_pData[0].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())

    def get_max_frames(self):
        return 1

    def get_plugin_pattern(self):
        return "PROJECTION"



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: dials_find_spots
   :platform: Unix
   :synopsis: A plugin to integrate azimuthally "symmetric" signals i.e. SAXS,\
       WAXS or XRD.Requires a calibration file

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
import numpy as np
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from scipy.ndimage import gaussian_filter

from dials.array_family import flex
from dials.algorithms.image.threshold import DispersionThreshold


@register_plugin
class DialsFindSpots(BaseFilter, CpuPlugin):
    """
    finding the single crystal peaks with dials
    :param spotsize: approximate maximum spot size. Default: 45.

    """

    def __init__(self):
        logging.debug("Starting DialsFindSpots")
        super(DialsFindSpots,
              self).__init__("DialsFindSpots")

    def process_frames(self, data):
        data = data[0]
        lp = gaussian_filter(data, 100)
        hp = data - lp # poormans background subtraction
        hp -= np.min(hp)
        sh = hp.shape
        hp = hp.astype('uint32')
        hp = flex.int(hp)
        
        mask = flex.bool(np.ones_like(hp).astype('bool'))
        result1 = flex.bool(np.zeros_like(hp).astype('bool'))
        spots = np.zeros_like(hp).astype('bool')
        
        for i in range(3, self.parameters['spotsize'], 5):
            algorithm = DispersionThreshold(sh, (i, i), 1, 1, 0, -1)
            #print type(hp), type(mask), type(result1)
            thing = algorithm(hp, mask, result1)
            spots = spots + result1.as_numpy_array()
        return [data, spots*data]

    def setup(self):
        in_datasets, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        diffractions = in_datasets[0]
        powder = out_datasets[0]
        single_crystal = out_datasets[1]
        powder.create_dataset(diffractions)
        single_crystal.create_dataset(diffractions)
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('DIFFRACTION', self.get_max_frames())
        out_pData[0].plugin_data_setup('DIFFRACTION', self.get_max_frames())
        out_pData[1].plugin_data_setup('DIFFRACTION', self.get_max_frames())

    def get_max_frames(self):
        return 'single'

    def nOutput_datasets(self):
        return 2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_template2
   :platform: Unix
   :synopsis: A template to create a plugin with a single input dataset and \
   multiple output datasets, that do not resemble the input dataset and are \
   not retained by the framework.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""
import numpy as np

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class PluginTemplate5(Plugin, CpuPlugin):
    """
    A plugin template with one in_dataset and two out_datasets that do not \
    resemble the in_dataset and are not retained by the framework, \
    e.g. vo_centering.

    :param example: Example of a plugin parameter. Default: None.
    :param preview: reduce the size of the data temporarily for this \
        plugin. Default: None.
    """

    def __init__(self):
        super(PluginTemplate5, self).__init__('PluginTemplate5')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 2

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()

        # get the full shape of the data before previewing
        full_shape = in_dataset[0].get_shape()

        # reduce the data as per data_subset parameter
        self.set_preview(in_dataset[0], self.parameters['preview'])

        # get the reduced shape of the data after previewing
        reduced_shape = in_dataset[0].get_shape()

        slice_dirs = np.array(in_dataset[0].get_slice_dimensions())
        new_shape = (np.prod(np.array(reduced_shape)[slice_dirs]), 1)
        full_length = (np.prod(np.array(full_shape)[slice_dirs]), 1)

        #=================== populate output datasets =========================
        # the output datasets are of a different type (i.e different shape,
        # axis labels and patterns)  to the input dataset, so more information
        # is required.

        # the first output dataset contains one value for each...
        out_dataset[0].create_dataset(shape=new_shape,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      remove=True,
                                      transport='hdf5')
        # currently there are no patterns assigned to this dataset - add one.
        out_dataset[0].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))

        # write something here about this being a dummy dataset...
        out_dataset[1].create_dataset(shape=full_length,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      remove=True,
                                      transport='hdf5')
        # currently there are no patterns assigned to this dataset - add one.
        out_dataset[1].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        #======================================================================

        #================== populate plugin datasets ==========================
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_max_frames())
        out_pData[0].plugin_data_setup('METADATA', 'single')
        out_pData[1].plugin_data_setup('METADATA', 'single')
        #======================================================================

    def pre_process(self):
        pass

    def process_frames(self, data):
        pass

    def post_process(self):
        # Add some information to post process *** how are these datasets
        # processed
        pass



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Segmentation by threhsolding based on lower and upper intensities
   :platform: Unix
   :synopsis: Segmentation by threhsolding based on lower and upper intensities

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

@register_plugin
class ThreshSegm(Plugin, CpuPlugin):
    """
    A Plugin to segment the data by providing two scalar values for lower and upper intensities

    :param min_intensity: A scalar to define lower limit for intensity, all values below are set to zero. Default: 0.
    :param max_intensity: A scalar to define upper limit for intensity, all values above are set to zero. Default: 0.01.
    :param value: An integer to set all values between min_intensity and max_intensity. Default: 1.
    :param pattern: pattern to apply this to. Default: "VOLUME_YZ".
    """

    def __init__(self):
        super(ThreshSegm, self).__init__("ThreshSegm")

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

    def pre_process(self):
        # extract given parameters
        self.min_limit = self.parameters['min_intensity']
        self.max_limit = self.parameters['max_intensity']
        self.value = self.parameters['value']

    def process_frames(self, data):
        thresh_result = np.uint8(np.zeros(np.shape(data[0])))
        thresh_result[(data[0] >= self.min_limit) & (data[0] < self.max_limit)] = self.value
        return thresh_result

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: plugin_runner
   :platform: Unix
   :synopsis: Plugin list runner, which passes control to the transport layer.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import logging

import savu.core.utils as cu
import savu.plugins.utils as pu
from savu.data.experiment_collection import Experiment


class PluginRunner(object):
    """ Plugin list runner, which passes control to the transport layer.
    """
    def __init__(self, options, name='PluginRunner'):
        class_name = "savu.core.transports." + options["transport"] \
                     + "_transport"
        cu.add_base(self, cu.import_class(class_name))
        super(PluginRunner, self).__init__()

        #  ********* transport function ***********
        self._transport_initialise(options)
        self.options = options
        # add all relevent locations to the path
        pu.get_plugins_paths()
        self.exp = Experiment(options)

    def _run_plugin_list(self):
        """ Create an experiment and run the plugin list.
        """
        self.exp._setup(self)

        plugin_list = self.exp.meta_data.plugin_list
        logging.info('Running the plugin list check')
        self._run_plugin_list_setup(plugin_list)

        exp_coll = self.exp._get_collection()
        n_plugins = plugin_list._get_n_processing_plugins()

        #  ********* transport function ***********
        logging.info('Running transport_pre_plugin_list_run()')
        self._transport_pre_plugin_list_run()

        cp = self.exp.checkpoint
        checkpoint_plugin = cp.get_checkpoint_plugin()
        for i in range(checkpoint_plugin, n_plugins):
            self.exp._set_experiment_for_current_plugin(i)
            memory_before = cu.get_memory_usage_linux()

            plugin_name = self.__run_plugin(exp_coll['plugin_dict'][i])

            self.exp._barrier(msg='PluginRunner: plugin complete.')

            memory_after = cu.get_memory_usage_linux()
            logging.debug("{} memory usage before: {} MB, after: {} MB, change: {} MB".format(
                plugin_name, memory_before, memory_after, memory_after - memory_before))

            #  ********* transport functions ***********
            # end the plugin run if savu has been killed
            if self._transport_kill_signal():
                self._transport_cleanup(i + 1)
                break
            self.exp._barrier(msg='PluginRunner: No kill signal... continue.')
            cp.output_plugin_checkpoint()

        #  ********* transport function ***********
        logging.info('Running transport_post_plugin_list_run')
        self._transport_post_plugin_list_run()

        # terminate any remaining datasets
        for data in list(self.exp.index['in_data'].values()):
            self._transport_terminate_dataset(data)

        self.__output_final_message()

        if self.exp.meta_data.get('email'):
            cu.send_email(self.exp.meta_data.get('email'))

        return self.exp

    def __output_final_message(self):
        kill = True if 'killsignal' in \
                       self.exp.meta_data.get_dictionary().keys() else False
        msg = "interrupted by killsignal" if kill else "Complete"
        stars = 40 if kill else 23
        cu.user_message("*" * stars)
        cu.user_message("* Processing " + msg + " *")
        cu.user_message("*" * stars)

    def __run_plugin(self, plugin_dict):
        plugin = self._transport_load_plugin(self.exp, plugin_dict)

        #  ********* transport function ***********
        self._transport_pre_plugin()
        cu.user_message("*Running the %s plugin*" % plugin.name)

        #  ******** transport 'process' function is called inside here ********
        plugin._run_plugin(self.exp, self)  # plugin driver

        self.exp._barrier(msg="Plugin returned from driver in Plugin Runner")
        cu._output_summary(self.exp.meta_data.get("mpi"), plugin)
        plugin._clean_up()
        finalise = self.exp._finalise_experiment_for_current_plugin()

        #  ********* transport function ***********
        self._transport_post_plugin()

        for data in finalise['remove'] + finalise['replace']:
            #  ********* transport function ***********
            self._transport_terminate_dataset(data)

        self.exp._reorganise_datasets(finalise)
        return plugin.name

    def _run_plugin_list_setup(self, plugin_list):
        """ Run the plugin list through the framework without executing the
        main processing.
        """
        plugin_list._check_loaders()
        self.__check_gpu()

        n_loaders = self.exp.meta_data.plugin_list._get_n_loaders()
        n_plugins = plugin_list._get_n_processing_plugins()
        plist = plugin_list.plugin_list

        # set loaders
        for i in range(n_loaders):
            pu.plugin_loader(self.exp, plist[i])
            self.exp._set_initial_datasets()

        # run all plugin setup methods and store information in experiment
        # collection
        count = 0
        for plugin_dict in plist[n_loaders:n_loaders + n_plugins]:
            self.__plugin_setup(plugin_dict, count)
            count += 1

        plugin_list._add_missing_savers(self.exp)

        #  ********* transport function ***********
        self._transport_update_plugin_list()

        # check added savers
        for plugin_dict in plist[n_loaders + count:]:
            self.__plugin_setup(plugin_dict, count)
            count += 1

        self.exp._reset_datasets()
        self.exp._finalise_setup(plugin_list)
        cu.user_message("Plugin list check complete!")


    def __plugin_setup(self, plugin_dict, count):
        self.exp.meta_data.set("nPlugin", count)
        plugin = pu.plugin_loader(self.exp, plugin_dict, check=True)
        plugin._revert_preview(plugin.get_in_datasets())
        plugin_dict['cite'] = plugin.get_citation_information()
        plugin._clean_up()
        self.exp._update(plugin_dict)
        self.exp._merge_out_data_to_in()

    def __check_gpu(self):
        """ Check if the process list contains GPU processes and determine if
        GPUs exists. Add GPU processes to the processes list if required."""
        if not self.exp.meta_data.plugin_list._contains_gpu_processes():
            return

        try:
            import pynvml as pv
        except Exception:
            logging.debug("pyNVML module not found")
            raise Exception("pyNVML module not found")

        try:
            pv.nvmlInit()
            count = int(pv.nvmlDeviceGetCount())
            if count == 0:
                raise Exception("No GPUs found")
            logging.debug("%s GPUs have been found.", count)

            if not self.exp.meta_data.get('test_state'):
                for i in range(count):
                    handle = pv.nvmlDeviceGetHandleByIndex(i)
                    if pv.nvmlDeviceGetComputeRunningProcesses(handle):
                        raise Exception("Unfortunately, GPU %i is busy. Try \
                            resubmitting the job to the queue." % i)
        except Exception as e:
            raise Exception("Unable to run GPU plugins: %s", str(e))
        self.__set_gpu_processes(count)

    def __set_gpu_processes(self, count):
        processes = self.exp.meta_data.get('processes')
        if not [i for i in processes if 'GPU' in i]:
            logging.debug("GPU processes missing. GPUs found so adding them.")
            cpus = ['CPU' + str(i) for i in range(count)]
            gpus = ['GPU' + str(i) for i in range(count)]
            for i in range(min(count, len(processes))):
                processes[processes.index(cpus[i])] = gpus[i]
            self.exp.meta_data.set('processes', processes)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: ortho_slice
   :platform: Unix
   :synopsis: A plugin render some slices from a volume and save them as images
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>
"""
from savu.plugins.driver.cpu_plugin import CpuPlugin

import os
import copy
import logging

import scipy.misc

import matplotlib.pyplot as plt

from savu.plugins.utils import register_plugin
from savu.plugins.plugin import Plugin

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('matplotlib')
logger.setLevel(logging.WARNING)


@register_plugin
class OrthoSlice(Plugin, CpuPlugin):
    """
    A plugin to calculate the centre of rotation using the Vo Method

    :u*param xy_slices: which XY slices to render. Default: 100.
    :u*param yz_slices: which YZ slices to render. Default: 100.
    :u*param xz_slices: which XZ slices to render. Default: 100.
    :u*param file_type: File type to save as. Default: 'png'.
    :u*param colourmap: Colour scheme to apply to the image. Default: 'magma'.
    :param out_datasets: Default out dataset names. Default: ['XY', 'YZ', 'XZ']
    """

    def __init__(self):
        super(OrthoSlice, self).__init__("OrthoSlice")

    def pre_process(self):
        self.image_path = \
            os.path.join(self.exp.meta_data.get('out_path'), 'OrthoSlice')
        if self.exp.meta_data.get('process') == 0:
            if not os.path.exists(self.image_path):
                os.makedirs(self.image_path)

    def process_frames(self, data):
        self.exp.log("XXXX Starting to run process_frames in Orthoslice")
        print("XXXX Starting to run process_frames in Orthoslice")

        in_dataset = self.get_in_datasets()

        fullData = in_dataset[0]

        ext = self.parameters['file_type']
        in_plugin_data = self.get_plugin_in_datasets()[0]
        pos = in_plugin_data.get_current_frame_idx()

        self.exp.log("frame position is %s" % (str(pos)))

        slice_info = [('xy_slices', 'VOLUME_XY'),
                      ('yz_slices', 'VOLUME_YZ'),
                      ('xz_slices', 'VOLUME_XZ')]

        colourmap = plt.get_cmap(self.parameters['colourmap'])

        # Set up the output list
        output_slices = []

        for direction, pattern in slice_info:
            # build the slice list
            slice_to_take = [slice(0)]*len(fullData.data.shape)

            # Fix the main slice dimentions that are outside the spatial dimentions
            slice_count = 0
            for i in list(self.slice_dims):
                slice_to_take[i] = slice(pos[slice_count], pos[slice_count]+1, 1)
                slice_count += 1

            # set all the spatial dimentions to get everything
            for i in list(self.spatial_dims):
                slice_to_take[i] = slice(None)

            # set the slice in the direction
            slice_value = self.parameters[direction]
            slice_to_take[self.axis_loc[pattern]] = slice(slice_value,
                                                          slice_value+1, 1)

            print(("Final slice is : %s of %s" % (str(slice_to_take),
                                                 str(fullData.data.shape))))
            self.exp.log("Final slice is : %s of %s" % (str(slice_to_take),
                                                        str(fullData.data.shape)))

            # now retreive the data from the fulldata given the current slice
            ortho_data = fullData.data[tuple(slice_to_take)].squeeze()
            output_slices.append(ortho_data)

            if ext is not 'None':
                image_data = ortho_data - ortho_data.min()
                image_data /= image_data.max()
                image_data = colourmap(image_data, bytes=True)

                filename = '%s_%03i_%s.%s' % (pattern, slice_value,
                                              str(pos), ext)
                self.exp.log("image-data shape is %s and filename is '%s'" %
                             (str(image_data.shape), filename))

                scipy.misc.imsave(os.path.join(self.image_path, filename), image_data)

        return output_slices


    def populate_meta_data(self, key, value):
        datasets = self.parameters['datasets_to_populate']
        in_meta_data = self.get_in_meta_data()[0]
        in_meta_data.set(key, value)
        for name in datasets:
            self.exp.index['in_data'][name].meta_data.set(key, value)

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()

        full_data_shape = list(in_dataset[0].get_shape())

        pattern = ['VOLUME_XY', 'VOLUME_YZ', 'VOLUME_XZ']
        fixed_axis = ['voxel_z', 'voxel_x', 'voxel_y']

        self.spatial_dims = []
        self.axis_loc = {}
        for patt, axis in zip(pattern, fixed_axis):
            self.axis_loc[patt] = in_dataset[0].get_data_dimension_by_axis_label(axis)
            self.spatial_dims.append(self.axis_loc[patt])

        self.all_dims = set(range(len(full_data_shape)))
        self.slice_dims = self.all_dims.difference(set(self.spatial_dims))

        self.exp.log("Axis Locations are %s" % (str(self.axis_loc)))

        # Sort out input data
        in_pData[0].plugin_data_setup('VOLUME_XY', self.get_max_frames())
        fixed_dim = in_dataset[0].get_data_dimension_by_axis_label('voxel_z')
        preview = [':']*len(in_dataset[0].get_shape())
        preview[fixed_dim] = str(0)
        self.set_preview(in_dataset[0], preview)

        # use this for 3D data (need to keep slice dimension)
        out_dataset = self.get_out_datasets()
        out_pData = self.get_plugin_out_datasets()

        for i, (p, axis) in enumerate(zip(pattern, fixed_axis)):
            fixed_dim = in_dataset[0].get_data_dimension_by_axis_label(axis)
            shape, patterns, labels = self._get_data_params(
                    in_dataset[0], full_data_shape, fixed_dim, p)
            out_dataset[i].create_dataset(axis_labels=labels,
                                          shape=tuple(shape),
                                          patterns=patterns)
            out_pData[i].plugin_data_setup(p, self.get_max_frames())

    def _get_data_params(self, data, full_data_shape, fixed_dim, p):
        shape = copy.copy(full_data_shape)
        if self.slice_dims:
            # for > 3D data
            del shape[fixed_dim]
            pattern = {data: ['%s.%i'%(p, fixed_dim)]}
            labels = {data: [str(fixed_dim)]}
        else:
            # for 3D data
            shape[fixed_dim] = 1
            pattern = {data: [p]}
            labels = data
        return shape, pattern, labels

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 3

    def get_max_frames(self):
        return 'single'

    def fix_transport(self):
        return 'hdf5'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ring_artefact_filter
   :platform: Unix
   :synopsis: A plugin to perform ring artefact removal

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from ccpi.reconstruction.parallelbeam import filters as ccpi_filters

from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.data.plugin_list import CitationInformation
from savu.plugins.filters.base_filter import BaseFilter


@register_plugin
class CcpiRingArtefactFilter(BaseFilter, CpuPlugin):
    """
    This plugin applies the same ring removal method as the DLS tomo_recon\
    reconstruction software.

    :u*param param_r: The correction strength - decrease (typically in order \
    of magnitude steps) to increase ring supression, or increase to reduce \
    ring supression. Default: 0.005.

    :param param_n: Unknown description (for plate-like objects \
    only). Default: 0.
    :u*param num_series: High aspect ration compensation (for plate-like \
    objects only) . Default: 1.
    """

    def __init__(self):
        super(CcpiRingArtefactFilter, self).__init__(
                "CcpiRingArtefactFilter")

    def process_frames(self, data):
        param_n = self.parameters['param_n']
        param_r = self.parameters['param_r']
        num_series = self.parameters['num_series']

        return ccpi_filters.aml_ring_artefacts(
                data[0], param_n, param_r, num_series)

    def get_plugin_pattern(self):
        return 'SINOGRAM'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The ring artefact removal algorithm used in this processing\n"
             "chain is taken from this work.")
        cite_info.bibtex = \
            ("@inproceedings{titarenko2010regularization,\n" +
             "title={Regularization methods for inverse problems in\n" +
             "X-ray tomography},\n" +
             "author={Titarenko, Valeriy and Bradley, Robert and Martin,\n" +
             "Christopher and Withers, Philip J and Titarenko, Sofya},\n" +
             "booktitle={Developments in X-Ray Tomography VII},\n" +
             "volume={7804},\n" +
             "pages={78040Z},\n" +
             "year={2010},\n" +
             "organization={International Society for Optics and Photonics}" +
             "}")
        cite_info.endnote = \
            ("%0 Conference Proceedings\n" +
             "%T Regularization methods for inverse problems in\n" +
             "X-ray tomography\n" +
             "%A Titarenko, Valeriy\n" +
             "%A Bradley, Robert\n" +
             "%A Martin, Christopher\n" +
             "%A Withers, Philip J\n" +
             "%A Titarenko, Sofya\n" +
             "%B Developments in X-Ray Tomography VII\n" +
             "%V 7804\n" +
             "%P 78040Z\n" +
             "%D 2010\n" +
             "%I International Society for Optics and Photonics")
        cite_info.doi = "doi: 10.1117/12.860260" 
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: template_loader
   :platform: Unix
   :synopsis: testing the template_loader plugin
.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import unittest
import tempfile
import shutil
import os
import numpy as np
import h5py as h5
import tifffile

from scripts.config_generator.content import Content
from collections import OrderedDict
from savu.test import test_utils as tu
from savu.plugins.loaders.utils import yaml_utils as yu
from savu.test.travis.framework_tests.plugin_runner_test import \
    run_protected_plugin_runner
import scripts.config_generator.config_utils as utils


class TemplateLoaderTest(unittest.TestCase):

    def test_1D_data(self):
        A = 11

        X = 10
        Y = 13

        expected_output_shape = (A, X, Y)
        expected_top_corners = np.arange(A, dtype=np.float32)

        self.create_N_tiffs(A, frame_size=(X, Y))
        # create a data file with the axis information
        self.data_file['entry/A'] = np.arange(A)
        self.data_file.close()

        # edit and save the yaml
        self.yaml['xrd']['data']['shape'] = '$(len(A_vals),)'
        self.yaml['xrd']['params'].update({'idx_A': 0,
                                           'idx_detx': 1,
                                           'idx_dety': 2,
                                           'A_vals': "$dfile['entry/A'][()]",
                                           'dims': "$range(0, 3)"})

        # add some axis labels
        self.yaml['xrd']['axis_labels'] = {0: {'dim': '$idx_A', 'name': 'A', 'value': '$A_vals', 'units': 'pixels'},
                                           1: self.detX_axis_label,
                                           2: self.detY_axis_label}

        self.save_yaml_and_change_process_list()
        run_protected_plugin_runner(tu.set_options(self.data_file_path,
                                                   process_file=self.process_list_path,
                                                   out_path=self.test_folder))

        # now check the result

        result = h5.File(os.path.join(self.test_folder, 'test_processed.nxs'), 'r')['entry/final_result_xrd/data']
        np.testing.assert_array_equal(result.shape, expected_output_shape,
                                      err_msg='The output shape is not as expected.')
        result_corners = result[..., -1, -1]
        np.testing.assert_equal(result_corners.dtype, expected_top_corners.dtype,
                                err_msg='The array does not output the correct type.')
        np.testing.assert_array_equal(result_corners, expected_top_corners,
                                      err_msg="The output values are not as expected")

    def test_again(self):
        A = 11

        X = 10
        Y = 13

        expected_output_shape = (A, X, Y)
        expected_top_corners = np.arange(A, dtype=np.float32)

        self.create_N_tiffs(A, frame_size=(X, Y))
        # create a data file with the axis information
        self.data_file['entry/A'] = np.arange(A)
        self.data_file.close()

        # edit and save the yaml
        self.yaml['xrd']['data']['shape'] = '$(len(A_vals),)'
        self.yaml['xrd']['params'].update({'idx_A': 0,
                                           'idx_detx': 1,
                                           'idx_dety': 2,
                                           'A_vals': "$dfile['entry/A'][()]",
                                           'dims': "$range(0, 3)"})

        # add some axis labels
        self.yaml['xrd']['axis_labels'] = {0: {'dim': '$idx_A', 'name': 'A', 'value': '$A_vals', 'units': 'pixels'},
                                           1: self.detX_axis_label,
                                           2: self.detY_axis_label}

        self.save_yaml_and_change_process_list()
        run_protected_plugin_runner(tu.set_options(self.data_file_path,
                                                   process_file=self.process_list_path,
                                                   out_path=self.test_folder))

        # now check the result

        result = h5.File(os.path.join(self.test_folder, 'test_processed.nxs'), 'r')['entry/final_result_xrd/data']
        np.testing.assert_array_equal(result.shape, expected_output_shape,
                                      err_msg='The output shape is not as expected.')
        result_corners = result[..., -1, -1]
        np.testing.assert_equal(result_corners.dtype, expected_top_corners.dtype,
                                err_msg='The array does not output the correct type.')
        np.testing.assert_array_equal(result_corners, expected_top_corners,
                                      err_msg="The output values are not as expected")

    def test_2D_data(self):
        A = 3
        B = 4

        X = 10
        Y = 13

        expected_output_shape = (A, B, X, Y)
        expected_top_corners = np.arange(A*B, dtype=np.float32).reshape((A, B))

        self.create_N_tiffs(A*B, frame_size=(X, Y))
        # create a data file with the axis information
        self.data_file['entry/A'] = np.arange(A)
        self.data_file['entry/B'] = np.arange(B)
        self.data_file.close()

        # edit and save the yaml
        self.yaml['xrd']['data']['shape'] = '$(len(A_vals), len(B_vals))'
        self.yaml['xrd']['params'].update({'idx_A': 0,
                                           'idx_B': 1,
                                           'idx_detx': 2,
                                           'idx_dety': 3,
                                           'A_vals': "$dfile['entry/A'][()]",
                                           'B_vals': "$dfile['entry/B'][()]",
                                           'dims': "$range(0, 4)"})

        self.yaml['xrd']['axis_labels'] = {0: {'dim': '$idx_A', 'name': 'A', 'value': '$A_vals', 'units': 'pixels'},
                                           1: {'dim': '$idx_B', 'name': 'B', 'value': '$B_vals', 'units': 'pixels'},
                                           2: self.detX_axis_label,
                                           3: self.detY_axis_label}

        self.save_yaml_and_change_process_list()

        run_protected_plugin_runner(tu.set_options(self.data_file_path,
                                                   process_file=self.process_list_path,
                                                   out_path=self.test_folder))

        result = h5.File(os.path.join(self.test_folder, 'test_processed.nxs'), 'r')['entry/final_result_xrd/data']
        np.testing.assert_array_equal(result.shape, expected_output_shape,
                                      err_msg='The output shape is not as expected.')
        result_corners = result[..., -1, -1]
        np.testing.assert_equal(result_corners.dtype, expected_top_corners.dtype,
                                err_msg='The array does not output the correct type.')
        np.testing.assert_array_equal(result_corners, expected_top_corners,
                                      err_msg="The output values are not as expected")

    def test_3D_data(self):
        A = 3
        B = 4
        C = 5

        X = 10
        Y = 13
        expected_output_shape = (A, B, C, X, Y)
        expected_top_corners = np.arange(A*B*C, dtype=np.float32).reshape((A, B, C))

        self.create_N_tiffs(A*B*C, frame_size=(X, Y))
        # create a data file with the axis information
        self.data_file['entry/A'] = np.arange(A)
        self.data_file['entry/B'] = np.arange(B)
        self.data_file['entry/C'] = np.arange(C)
        self.data_file.close()

        # edit and save the yaml
        self.yaml['xrd']['data']['shape'] = '$(len(A_vals), len(B_vals), len(C_vals))'
        self.yaml['xrd']['params'].update({'idx_A': 0,
                                           'idx_B': 1,
                                           'idx_C': 2,
                                           'idx_detx': 3,
                                           'idx_dety': 4,
                                           'A_vals': "$dfile['entry/A'][()]",
                                           'B_vals': "$dfile['entry/B'][()]",
                                           'C_vals': "$dfile['entry/C'][()]",
                                           'dims': "$range(0, 4)"})

        self.yaml['xrd']['axis_labels'] = {0: {'dim': '$idx_A', 'name': 'A', 'value': '$A_vals', 'units': 'pixels'},
                                           1: {'dim': '$idx_B', 'name': 'B', 'value': '$B_vals', 'units': 'pixels'},
                                           2: {'dim': '$idx_C', 'name': 'C', 'value': '$C_vals', 'units': 'pixels'},
                                           3: self.detX_axis_label,
                                           4: self.detY_axis_label}

        self.save_yaml_and_change_process_list()
        run_protected_plugin_runner(tu.set_options(self.data_file_path,
                                                   process_file=self.process_list_path,
                                                   out_path=self.test_folder))

        result = h5.File(os.path.join(self.test_folder, 'test_processed.nxs'), 'r')['entry/final_result_xrd/data']
        np.testing.assert_array_equal(result.shape, expected_output_shape,
                                      err_msg='The output shape is not as expected.')
        result_corners = result[..., -1, -1]
        np.testing.assert_equal(result_corners.dtype, expected_top_corners.dtype,
                                err_msg='The array does not output the correct type.')
        np.testing.assert_array_equal(result_corners, expected_top_corners,
                                      err_msg="The output values are not as expected")

    def setUp(self):
        self.test_folder = tempfile.mkdtemp(suffix='template_test/')
        self.tif_folder = os.path.join(self.test_folder, 'tiffs/')
        os.mkdir(self.tif_folder)

        # copy across the process list to the working directory
        self.process_list_path = os.path.join(self.test_folder, 'xrd_template_test.nxs')
        shutil.copyfile(tu.get_test_process_path('xrd_template_test.nxs'),
                        self.process_list_path)

        self.detX_axis_label = {'dim': '$idx_detx', 'name': 'detector_x', 'value': None, 'units': 'pixels'}
        self.detY_axis_label = {'dim': '$idx_dety', 'name': 'detector_y', 'value': None, 'units': 'pixels'}

        self.yaml = OrderedDict()
        # now make some standard modifications
        self.yaml['inherit'] = [tu.get_test_data_path(os.path.join('i18_templates', 'xrd_calibration.yml'))]
        self.yaml['xrd'] = OrderedDict()
        self.yaml['xrd']['params'] = {}
        self.yaml['xrd']['data'] = {}
        self.yaml['xrd']['data']['folder'] = self.tif_folder
        self.yaml['xrd']['params']['cfile'] = \
            "$h5py.File('%s', 'r')" % tu.get_test_data_path('LaB6_calibration_new.nxs')

        self.yaml['xrd']['patterns'] = {}
        self.yaml['xrd']['patterns']['DIFFRACTION'] = {'core_dims': '$(idx_detx, idx_dety)',
                                                       'slice_dims': '$tuple([d for d in dims if d not in [idx_detx, idx_dety]])'}
        self.yaml['xrd']['axis_labels'] = {}
        self.yaml['xrd']['metadata'] = {}
        self.data_file_path = os.path.join(self.test_folder, 'test_data.nxs')
        self.data_file = h5.File(self.data_file_path, 'w')  # this will have the axes in.

    def tearDown(self):
        shutil.rmtree(self.test_folder, ignore_errors=True)

    def create_N_tiffs(self, N, file_pattern='%04d.tif', frame_size=(10, 13)):
        '''
        Creates a stack of N tiffs with given frame_size and file pattern in self.tif_folder.
        :param N: The number of tiffs in the stack
        :param file_pattern: The printf pattern for the file names.
        :param frame_size: The shape of the frames. Default: (10,13).
        '''

        data = np.ones(frame_size, dtype=np.float)
        for idx in range(N):
            file_path = os.path.join(self.tif_folder, (file_pattern % idx))
            tifffile.imsave(file_path, data*idx)

    def save_yaml_and_change_process_list(self):
        self.yml_path = os.path.join(self.test_folder, 'xrd_tiff.yml')
        with open(self.yml_path, 'w') as f:
            yu.dump_yaml(self.yaml, f)
        # now set the template path in the process list
        # uses the parameter name instead of number
        # because the number can change between runs
        self.process_list = Content()
        self.process_list.fopen(self.process_list_path, update=False)
        self.process_list.modify('1', 'yaml_file', self.yml_path)
        self.process_list.save(self.process_list_path)


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_transport_data
   :platform: Unix
   :synopsis: Base class for transport mechanism data functions.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy
import logging
import numpy as np
from typing import List


class BaseTransportData(object):
    """
    Implements functions associated with the transport of the data.
    """

    def __init__(self, data_obj, name='BaseTransportData'):
        self.data = data_obj
        self.mfp = None

    def _get_data_obj(self):
        return self.data

    def _get_slice_list_per_process(self, expInfo):
        """
        A slice list required by the current process.
        """
        raise NotImplementedError("get_slice_list_per_process needs to be"
                                  " implemented in  %s", self.__class__)

    def _get_padded_data(self, input_slice_list):
        """
        Fetch the data with relevant padding (as determined by the plugin).
        """
        raise NotImplementedError("_get_padded_data needs to be"
                                  " implemented in  %s", self.__class__)

    def _calc_max_frames_transfer(self, nFrames):
        """ Calculate the number of frames to transfer from file at a time.
        """
        raise NotImplementedError("_calc_max_frames_transfer needs to be"
                                  " implemented in  %s", self.__class__)

    def _calc_max_frames_process(self, nFrames):
        nSlices = self.params['shape'][self.params['sdir'][0]]
        mfp = 1 if nFrames == 'single' else nFrames if \
            isinstance(nFrames, int) else self.mfp if self.mfp else nSlices
        return int(mfp)

    def _calc_max_frames_transfer_single(self, nFrames):
        """ Only one transfer per process """
        self.params = self.data._get_plugin_data().meta_data.get_dictionary()
        mft = np.ceil(
                self.params['total_frames']/float(self.params['mpi_procs']))

        # added for basic runner
        nSlices = self.params['shape'][self.params['sdir'][0]]
        self.mfp = nFrames if isinstance(nFrames, int) else min(mft, nSlices)

        fchoices, size_list = self._get_frame_choices(self.params['sdir'], mft)
        self.mft = mft
        return int(mft), size_list[fchoices.index(mft)]

    def _calc_max_frames_transfer_multi(self, nFrames):
        """ Multiple transfer per process """
        self.params = self.data._get_plugin_data().meta_data.get_dictionary()
        mft, fchoices, size_list = self.__get_optimum_distribution(nFrames)
        if nFrames == 'single':
            return mft, size_list[fchoices.index(mft)]
        nSlices = self.params['shape'][self.params['sdir'][0]]
        self.mfp = nFrames if isinstance(nFrames, int) else min(mft, nSlices)
        mft, fchoices, size_list = \
            self.__refine_distribution_for_multi_mfp(mft, size_list, fchoices)
        self.mft = mft
        return mft, size_list[fchoices.index(mft)]

    def _set_boundaries(self):
        b_per_f = self.params.get('bytes_per_frame')
        b_per_p = self.params.get('bytes_per_process')

        mem_multiply = \
            self._get_data_obj()._get_plugin_data()._plugin.get_mem_multiply()

        settings = self.data.exp.meta_data.get(
                ['system_params', 'data_transfer_settings'])
        max_bytes = self.__convert_str(
                settings['max_bytes'], b_per_p)*mem_multiply
        bytes_threshold = self.__convert_str(
                settings['bytes_threshold'], b_per_p)*mem_multiply
        b_per_p = b_per_p if b_per_p < bytes_threshold else bytes_threshold

        min_bytes = self.__convert_str(settings['min_bytes'], b_per_p)     
        max_mft = int(np.floor(float(max_bytes)/b_per_f))
        if max_mft == 0:
            raise Exception("The size of a single frame exceeds the permitted "
                            "maximum bytes per frame.")
        min_mft = int(max(np.floor(float(min_bytes) / b_per_f), 1))

        return min_mft, max_mft

    def __convert_str(self, val, b_per_p):
        if isinstance(val, str):
            # FIXME this is still a potential security risk!
            value = eval(val, {'__builtins__': None, 'b_per_p': b_per_p})
            return value

    def __get_optimum_distribution(self, nFrames):
        """ The number of frames each process should retrieve from file at a
        time.
        """
        min_mft, max_mft = self.__get_boundaries(nFrames)

        if isinstance(nFrames, int) and nFrames > max_mft:
            logging.warning("The requested %s frames excedes the maximum "
                         "preferred of %s." % (nFrames, max_mft))
            max_mft = nFrames

        # find all possible choices of nFrames, being careful with boundaries
        sdir = self.params['sdir']
        # for multiple mft slice dimensions
        #slice_shape = [self.params['shape'][d] for d in sdir]
        #nSlices = np.prod(slice_shape)
        nSlices = self.params['shape'][sdir[0]]

        nFrames = 1 if isinstance(nFrames, str) else nFrames
        fchoices, size_list = self._get_frame_choices(
            sdir, nFrames, min(max_mft, max(nFrames, nSlices)))

        threshold_idx = [i for i in range(len(fchoices)) if fchoices[i] >= min_mft]

        if threshold_idx:
            fchoices = [fchoices[i] for i in threshold_idx]
            size_list = [size_list[i] for i in threshold_idx]

# use this for multiple mft slice dimensions
#        mft, idx = self._find_best_frame_distribution(
#            size_list, slice_shape, self.params['mpi_procs'],
#            idx=True)
        mft, idx = self._find_best_frame_distribution(
            fchoices, nSlices, self.params['mpi_procs'], idx=True)
        return int(mft), fchoices, size_list

    def __refine_distribution_for_multi_mfp(self, mft, size_list, fchoices):
        flimit = self._get_data_obj()._get_plugin_data().get_frame_limit()
        mfp = self.mfp
        if flimit and flimit < mfp:
            mfp = flimit

        multi = self._find_multiples_of_b_that_divide_a(mft, mfp)
        possible = sorted(list(set(set(multi).intersection(set(fchoices)))),
                          reverse=True)
        idx = [fchoices.index(i) for i in possible]
        possible_size_list = [size_list[i] for i in idx]

        # closest of fchoices to mfp plus difference as boundary padding
        if not possible:
            mft, _ = self._find_closest_lower(fchoices[::-1], mfp)  # does this need to be updated?
        else:
            size_list = possible_size_list
            fchoices = possible
# use this for multiple mft slice dimensions
#            mft, idx = self._find_best_frame_distribution(
#                    size_list, self.params['total_frames'],
#                    self.params['mpi_procs'], idx=True)
            mft, idx = self._find_best_frame_distribution(
                    possible, self.params['total_frames'],
                    self.params['mpi_procs'], idx=True)

        self.mfp = mfp
        return int(mft), fchoices, size_list

    def __get_boundaries(self, nFrames):
        min_mft, max_mft = self._set_boundaries()
        if isinstance(nFrames, int) and nFrames > max_mft:
            logging.warning("The requested %s frames excedes the maximum "
                         "preferred of %s." % (nFrames, max_mft))
            max_mft = nFrames
        return min_mft, max_mft

    def _get_slice_dir_index(self, dim, boolean=False):
        starts, stops, steps, chunks = \
            self.data.get_preview().get_starts_stops_steps()
        if chunks[dim] > 1:
            dir_idx = np.ravel(np.transpose(self._get_slice_dir_matrix(dim)))
            if boolean:
                return self.__get_bool_slice_dir_index(dim, dir_idx)
            return dir_idx
        else:
            fix_dirs, value = \
                self.data._get_plugin_data()._get_fixed_dimensions()
            if dim in fix_dirs:
                return value[fix_dirs.index(dim)]
            else:
                return np.arange(starts[dim], stops[dim], steps[dim])

    def __get_bool_slice_dir_index(self, dim, dir_idx):
        shape = self.data_info.get('orig_shape')[dim]
        bool_idx = np.ones(shape, dtype=bool)
        bool_idx[dir_idx] = True
        return bool_idx

    def _get_slice_dir_matrix(self, dim):
        starts, stops, steps, chunks = \
            self.data.get_preview().get_starts_stops_steps()
        chunk = chunks[dim]
        a = np.tile(np.arange(starts[dim], stops[dim], steps[dim]), (chunk, 1))
        b = np.transpose(np.tile(np.arange(chunk)-chunk // 2, (a.shape[1], 1)))
        dim_idx = a + b
        if dim_idx[dim_idx < 0].size:
            raise Exception('Cannot have a negative value in the slice list.')
        return dim_idx

    def _find_closest_lower(self, vlist, value):
        rem = [f if f != 0 else value for f in [m % value for m in vlist]]
        min_val = min(rem, key=lambda x: abs(x-value))
        idx = rem.index(min_val)
        return vlist[idx], idx

# use this for multiple mft slice dimensions
#    def _get_frame_choices(self, sdir, max_mft):
#        """ Find all possible combinations of increasing slice dimension sizes
#        with their product less than max_mft and return a list of these
#        products. """
#        nDims = len(sdir)
#        temp = [1]*len(sdir)
#        shape = self.data.get_shape()
#        idx = 0
#        choices = []
#        size_list = []
#
#        while(np.prod(temp) <= max_mft):
#            dshape = shape[sdir[idx]]
#            choices.append(np.prod(temp))
#            size_list.append(copy.copy(temp))
#
#            if temp[idx] == dshape:
#                idx += 1
#                if idx == nDims:
#                    break
#            temp[idx] += 1
#
#        return choices[::-1], size_list[::-1]
# the above method is commented out as it was found to be slower than the one
# below - try again in a different version of HDF5

    def _get_frame_choices(self, sdir, min_mft, max_mft):
        """ Find all possible combinations of increasing slice dimension sizes
        with their product less than max_mft and return a list of these
        products. """
        temp = [1]*len(sdir)
        #temp[0] = min_mft # this works for ccpi but not mpimap
        shape = self.data.get_shape()
        idx = 0
        choices = []
        size_list = []

        while(np.prod(temp) <= max_mft):
            dshape = shape[sdir[idx]]
            choices.append(np.prod(temp))
            size_list.append(copy.copy(temp))

            if temp[idx] == dshape:
                break
            temp[idx] += 1

        return choices[::-1], size_list[::-1]

    def _find_multiples_of_b_that_divide_a(self, a, b):
        """ Find all positive multiples of b that divide a. """
        val_list = []
        i = 0
        val = (int(a/b)+i)*b
        while(val > 0):
            if a % val == 0:
                val_list.append(val)
            i -= 1
            val = (int(a/b)+i)*b
        return val_list

    def _find_best_frame_distribution(self, flist, nframes, nprocs, idx=False):
        """ Determine which of the numbers in the list of possible frame
        chunks gives the best distribution of frames per process. """
        multi_list: List[float] = [(nframes / float(v)) / nprocs for v in flist]
        min_val, closest_lower_idx = self._find_closest_lower(multi_list, 1)
        if idx:
            return flist[closest_lower_idx], closest_lower_idx
        return flist[closest_lower_idx]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_checkpoint_test
   :platform: Unix
   :synopsis: A base class from which checkpoint tests inherit.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import tempfile
import h5py
import unittest
import numpy as np
from shutil import copyfile
import shutil

from savu.core.plugin_runner import PluginRunner
from savu.core.checkpointing import Checkpointing
from savu.core.utils import ensure_string
from savu.data.experiment_collection import Experiment


class BaseCheckpointTest(object):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        cp = self._get_checkpoint(cfile=False)
        folder = os.path.join(self.tmpdir, 'checkpoint')
        fname = 'process0' + cp._filename
        self.cfile = os.path.join(folder, fname)
        options = self.get_options()
        exp = self._run_plugin_list(options)
        plist = self.set_plist(exp)

        # copy original nexus file
        nxs_file = exp.meta_data.get('nxs_filename')
        copy_nxs = ''.join(nxs_file.split('.nxs')[0:-1]) + '_copy.nxs'
        copyfile(nxs_file, copy_nxs)

        # create a checkpoint file with fixed values
        cp = self._get_checkpoint()
        cp._file = self.cfile
        cp._initialise(None)
        self.set_global_parameters(options, plist, nxs_file, copy_nxs, cp)

    def tearDown(self):
        cp_folder = os.path.join(self.tmpdir, 'checkpoint')
        #self._empty_folder(cp_folder)
        #os.removedirs(cp_folder)
        #self._empty_folder(self.tmpdir)
        #os.removedirs(self.tmpdir)
        shutil.rmtree(cp_folder, ignore_errors=True)
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def _empty_folder(self, folder):
        for f in os.listdir(folder):
            file_path = os.path.join(folder, f)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception as e:
                print(e)

    def set_global_parameters(self, options, plist, nxs_file, copy_nxs, cp):
        self.options = options
        self.plist = plist
        self.nxs_file = nxs_file
        self.orig_nxs = copy_nxs
        self.cp = cp

    def get_options(self, cfile=True):
        options = self.set_data_options()
        options['out_path'] = self.tmpdir
        options['nxs_filename'] = self.cfile if cfile else None
        options['processes'] = list(range(1))
        options['process'] = 0
        options['mpi'] = False
        options['inter_path'] = self.tmpdir
        return options

    def _get_checkpoint(self, cfile=True):
        exp = Experiment(self.get_options(cfile=cfile))
        checkpoint = Checkpointing(exp)
        return checkpoint

    def _run_plugin_list(self, options):
        plugin_runner = PluginRunner(options)
        exp = plugin_runner._run_plugin_list()
        return exp

    def test_checkpoint_1(self):
        # 1: Killed before checkpoint file created

        # No checkpoint file
        os.remove(self.cfile)
        self._refresh_nxs_file()
        vals = (0, 0, 0, 'plugin')
        with self.assertRaises(Exception) as context:
            self._rerun_from_checkpoint(*vals)
        self.assertTrue('No checkpoint file found.', context.exception)

    def test_plugin_level_2(self):
        # 2: Killed before checkpoint file fully populated (initialised)
        # Remove two datasets from checkpoint file
        vals = (0, None, None, 'plugin')
        self._amend_checkpoint_file(0, None, None)
        self._refresh_nxs_file()
        self._rerun_from_checkpoint(*vals)

        # Remove one dataset from checkpoint file
        vals = (0, 1, 0, 'subplugin')
        self._amend_checkpoint_file(0, 1, None)
        self._refresh_nxs_file()
        self._rerun_from_checkpoint(*vals)

        # Remove all datasets from checkpoint
        vals = (0, 0, 0, 'plugin')
        self._amend_checkpoint_file(None, None, None)
        self._refresh_nxs_file()
        self._rerun_from_checkpoint(*vals)

    def test_plugin_level_3(self):
        # 3: Killed before processing started

        vals = (0, 0, 0)
        self._amend_checkpoint_file(*vals)
        self._refresh_nxs_file()
        self._rerun_from_checkpoint(*(vals + ('plugins',)))

    def test_plugin_level_4a(self):
        # 4: Killed during processing
        #    a) standard
        vals = (3, 0, 0, 'plugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

        os.remove(self.nxs_file)
        vals = (3, 3, 3, 'subplugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

    def test_plugin_level_4b(self):
        # 4: Killed during processing
        #    b) after a plugin that doesn't populate nxs file - (works with full-field processing list)
        vals = (5, 3, 3, 'plugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

        os.remove(self.nxs_file)
        vals = (5, 3, 3, 'subplugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

    def test_plugin_level_4c(self):
        # 4: Killed during processing
        #    c) re-loading of a different data type, e.g., ImageKey
        vals = (1, 3, 3, 'plugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

        os.remove(self.nxs_file)
        vals = (1, 3, 3, 'subplugin')
        self._set_checkpoint_parameters(*vals)
        # rerun from checkpoint
        self._rerun_from_checkpoint(*vals)

    def _set_checkpoint_parameters(self, p_no, tidx, pidx, level):
        self._amend_nxs_file(self.orig_nxs, self.nxs_file, self.plist, p_no)
        self._amend_checkpoint_file(p_no, tidx, pidx)

    def _rerun_from_checkpoint(self, p_no, tidx, pidx, level, options=None):
        options = self.get_options() if options is None else options
        options['checkpoint'] = level
        exp = self._run_plugin_list(options)
        self._assert_checkpoint_params_equal(p_no, tidx, pidx, level, exp)
        self._assert_entries_equal(self.orig_nxs, self.nxs_file)

    def _refresh_nxs_file(self, entry=True):
        with h5py.File(self.nxs_file, 'w') as f:
            if entry:
                f.create_group('entry')

    def _amend_nxs_file(self, copy_file, nxs_file, plist, p_no):
        # update nxs file and checkpoint file
        with h5py.File(copy_file, 'r') as f:
            datasets = self._read_nexus_file(f, [])
            datasets = [d for d in datasets for p in
                        self._flatten(plist[0:p_no], []) if p in d.name]
            with h5py.File(nxs_file, 'w') as n:
                for d in datasets:
                    group = n.require_group(d.parent.name)
                    f.copy(d, group)

    def _flatten(self, alist, flat_list=[]):
        for l in alist:
            if isinstance(l, list):
                self._flatten(l, flat_list=flat_list)
            else:
                flat_list.append(l)
        return flat_list

    def _amend_checkpoint_file(self, p_no, tidx, pidx):
        with h5py.File(self.cfile, 'a') as f:
            self._create_dataset(f, 'transfer_idx', tidx)
            self._create_dataset(f, 'process_idx', pidx)
            self._create_dataset(f, 'completed_plugins', p_no)

    def _create_dataset(self, f, name, data):
        if name in list(f.keys()):
            f.__delitem__(name)
        if data is not None:
            f.create_dataset(name, data=data, dtype=np.int16)

    def _read_nexus_file(self, nxsfile, datasets):
        # find NXdata
        for key, value in list(nxsfile.items()):
            if self._is_nxdata(value):
                datasets.append(value)
            elif isinstance(value, h5py.Group) and key != 'input_data':
                self._read_nexus_file(value, datasets)
        return datasets

    def _is_nxdata(self, value):
        check = 'NX_class' in list(value.attrs.keys()) and ensure_string(value.attrs['NX_class']) == 'NXdata'
        return check

    def _assert_checkpoint_params_equal(self, p_no, tidx, pidx, level, exp):
        # find the checkpoint parameters and assert equal
        cp = exp.checkpoint
        used_p_no, used_tidx, used_pidx = cp.get_start_values()

        if level == 'plugin':
            tidx = 0
            pidx = 0

        self.assertEqual(used_tidx, tidx)
        self.assertEqual(used_pidx, pidx)
        self.assertEqual(used_p_no, p_no)

    def _assert_entries_equal(self, f1, f2):
        # get final entries in both files and check that they match
        with h5py.File(f1, 'r') as f1, h5py.File(f2, 'r') as f2:
            names = [d.name for d in self._read_nexus_file(f1, [])]
            final_res = [n for n in names if 'final_result' in n]
            for res in final_res:
                data = '/'.join((res, 'data'))
                np.testing.assert_array_almost_equal(f1[data][:], f2[data][:])

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: yaml_loader
   :platform: Unix
   :synopsis: A class to load data from a non-standard nexus/hdf5 file using \
   descriptions loaded from a yaml file.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import h5py
import fnmatch
import difflib
import numpy as np

import savu.core.utils as cu
from savu.plugins.utils import register_plugin
from savu.plugins.loaders.yaml_converter import YamlConverter
from savu.data.data_structures.data_types.stitch_data import StitchData


@register_plugin
class Hdf5TemplateLoader(YamlConverter):
    """
    A class to load data from a non-standard nexus/hdf5 file using \
    descriptions loaded from a yaml file.

    :u*param yaml_file: Path to the file containing the data \
        descriptions. Default: None.
    """

    def __init__(self, name='Hdf5TemplateLoader'):
        super(Hdf5TemplateLoader, self).__init__(name)

    def set_data(self, dObj, data):
        path = data['path'] if 'path' in list(data.keys()) else None
        if not path:
            emsg = 'Please specify the path to the data in the h5 file.'
            raise Exception(emsg)

        file_path = self.exp.meta_data.get("data_file") if 'file' not in \
            list(data.keys()) else data['file']
        file_path = self.update_value(dObj, file_path)
        dObj.backing_file = h5py.File(file_path, 'r')

        basename = os.path.basename(path)
        if len(basename.split('*')) > 1:
            return self._stitch_data(dObj, path, data)
        return self._setup_data(dObj, path)

    def _stitch_data(self, dObj, path, data):
        stype, dim = self._get_stitching_info(data)
        remove = data['remove'] if 'remove' in list(data.keys()) else None

        group_name, data_name = os.path.split(path)
        # find all files with the given name
        group = dObj.backing_file.require_group(group_name)
        matches = fnmatch.filter(list(group.keys()), data_name)

        number = []
        for m in matches:
            diff_number = ''
            for diff in difflib.ndiff(m, data_name):
                split = diff.split('- ')
                if len(split) > 1:
                    diff_number += split[-1]
            number.append(int(diff_number))

        matches = [matches[i] for i in np.argsort(number)]
        dObj.data_info.set('wildcard_values', sorted(number))

        data_obj_list = []
        for match in matches:
            match_path = os.path.join(group_name, match)
            sub_obj = self.exp.create_data_object('in_data', match)
            sub_obj.backing_file = dObj.backing_file
            data_obj_list.append(self._setup_data(sub_obj, match_path))
            del self.exp.index['in_data'][match]

        if data_obj_list:
            dObj.data = StitchData(data_obj_list, stype, dim, remove=remove)
            dObj.set_original_shape(dObj.data.get_shape())
        else:
            cu.user_message("The data set %s is empty." % data_name)

        return dObj

    def _get_stitching_info(self, data):
        if 'stack' in list(data.keys()):
            return 'stack', data['stack']
        elif 'cat' in list(data.keys()):
            return 'cat', data['cat']
        else:
            msg = 'Please specify the dimension to stack or concatenate.'
            raise Exception(msg)

    def _setup_data(self, dObj, path):
        dObj.data = dObj.backing_file[self.update_value(dObj, path)]
        dObj.set_shape(dObj.data.shape)
        return dObj



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: map_3dto4d_h5
   :platform: Unix
   :synopsis: A module for loading a 3D hdf5 dataset that contains a set of \
       4D scans.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.data.data_structures.data_types.base_type import BaseType


class Map3dto4dh5(BaseType):
    """ This class converts a 3D dataset to a 4D dataset. """

    def __init__(self, data_obj, n_angles):
        self.data_obj = data_obj
        self.data = data_obj.data
        self.n_angles = n_angles
        shape = self.data.shape
        super(Map3dto4dh5, self).__init__()

        import inspect
        if inspect.isclass(type(self.data)):
            self.add_base_class_with_instance(type(self.data), self.data)

        new_shape = (n_angles, shape[1], shape[2], shape[0] // n_angles)
        self.shape = new_shape

    def clone_data_args(self, args, kwargs, extras):
        args = ['self', 'n_angles']
        return args, kwargs, extras

    def __getitem__(self, idx):
        rot_dim = \
            self.data_obj.get_data_dimension_by_axis_label('rotation_angle')
        n_angles = self.shape[rot_dim]
        idx_dim3 = np.arange(idx[3].start, idx[3].stop, idx[3].step)
        idx_dim0 = np.arange(idx[0].start, idx[0].stop, idx[0].step)
        idx_dim0 = np.ravel(idx_dim3.reshape(-1, 1)*n_angles + idx_dim0)

        size = [len(np.arange(i.start, i.stop, i.step)) for i in idx]
        data = np.empty(size)

        change = np.where(idx_dim0[:-1] // n_angles != idx_dim0[1:] // n_angles)[0]
        start = idx_dim0[np.append(0, change+1)]
        stop = idx_dim0[np.append(change, len(idx_dim0)-1)] + 1
        length = stop - start

        for i in range(len(start)):
            new_slice = (slice(start[i], stop[i], idx[0].step), idx[1], idx[2])
            data[0:length[i], :, :, i] = self.data[new_slice]
        return data

    def get_shape(self):
        return self.shape



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: An Example median_3x3_filter
   :platform: Unix
   :synopsis: A plugin to filter each frame with a 3x3 median filter

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import logging

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin

import scipy.signal.signaltools as sig


class ExampleMedianFilter(BaseFilter, CpuPlugin):
    """
    A plugin to filter each frame with a 3x3 median filter

    :param kernel_size: Kernel size for the filter. Default: (1, 3, 3).
    """

    def __init__(self):
        logging.debug("Starting Median Filter")
        super(ExampleMedianFilter, self).__init__("ExampleMedianFilter")

    def filter_frames(self, data):
        result = sig.medfilt(data[0], self.parameters['kernel_size'])
        return result

    def set_filter_padding(self, in_data, out_data):
        padding = (self.parameters['kernel_size'][0]-1) // 2
        in_data[0].padding = {'pad_multi_frames': padding}
        out_data[0].padding = {'pad_multi_frames': padding}

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 8



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxstxm_loader
   :platform: Unix
   :synopsis: A class for loading nxstxm data

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.base_loader import BaseLoader
import numpy as np
from savu.plugins.utils import register_plugin
import logging
import h5py

@register_plugin
class TxmLoader(BaseLoader):
    """
    A class to load tomography data from the txm
    """

    def __init__(self, name='TxmLoader'):
        super(TxmLoader, self).__init__(name)

    def setup(self):
        """
         Define the input nexus file

        :param path: The full path of the NeXus file to load.
        :type path: str
        """

        exp = self.exp
        data_obj = exp.create_data_object('in_data', 'tomo')

        data_obj.backing_file = \
            h5py.File(self.exp.meta_data.get("data_file"), 'r')

        data_obj.data = data_obj.backing_file['/entry1/data']
        sh = data_obj.data.shape
        #print sh
        ### set the rotation
        rotation_angle = \
            data_obj.backing_file['entry1/theta'][()].astype(float)
        #print rotation_angle.shape
        data_obj.meta_data.set('rotation_angle', rotation_angle)
        data_obj.set_axis_labels('rotation_angle.degrees',
                                 'detector_y.pixel',
                                 'detector_x.pixel')

        data_obj.add_pattern('PROJECTION', core_dims=(1, 2),
                             slice_dims=(0,))
        data_obj.add_pattern('SINOGRAM', core_dims=(0, 2),
                             slice_dims=(1,))

        data_obj.set_shape(sh)
        self.set_data_reduction_params(data_obj)

        #print data_obj.get_data_patterns()
